from lona.html.abstract_node import AbstractNode
from lona.html.widget_data import WidgetData
from lona.html.node_list import NodeList
from lona.protocol import NODE_TYPE


class Widget(AbstractNode):
    FRONTEND_WIDGET_CLASS = ''

    @property
    def nodes(self):
        if not hasattr(self, '_nodes'):
            self._nodes = NodeList(self)

        return self._nodes

    @nodes.setter
    def nodes(self, value):
        if not hasattr(self, '_nodes'):
            self._nodes = NodeList(self)

        self._nodes._reset(value)

    @property
    def data(self):
        if not hasattr(self, '_data'):
            self._data = WidgetData(self)

        return self._data

    @data.setter
    def data(self, value):
        if not hasattr(self, '_data'):
            self._data = WidgetData(self)

        self._data._reset(value)

    def __len__(self):
        return self._nodes.__len__()

    # node helper #############################################################
    def remove(self):
        if not self.parent:
            raise RuntimeError('node has no parent node')

        self.parent.remove(self)

    # serialization ###########################################################
    def _serialize(self):
        return [
            NODE_TYPE.WIDGET,
            self.id,
            self.FRONTEND_WIDGET_CLASS,
            self.nodes._serialize(),
            self.data._serialize(),
        ]

    # string representation ###################################################
    def __str__(self):
        return f'<!--lona-widget:{self.id}-->\n{self.nodes}\n<!--end-lona-widget:{self.id}-->'

    def __repr__(self):
        return self.__str__()

    # node helper #############################################################
    def hide(self):
        with self.lock:
            for node in self.nodes:
                node.hide()

    def show(self):
        with self.lock:
            for node in self.nodes:
                node.show()

    def set_text(self, text):
        self.nodes = [str(text)]

    def get_text(self):
        with self.lock:
            text = []

            for node in self.nodes:
                text.append(node.get_text())

            return ' '.join(text).strip()



from pprint import pformat

from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table


class LonaSettingsCommand:
    """
    List currently loaded Lona settings
    """

    NAME = 'lona_settings'

    def __init__(self, repl):
        self.repl = repl

    def complete(self, text, state, line_buffer):
        names = sorted(self.repl.locals['server'].settings)
        candidates = []

        for name in names:
            if name.startswith(text):
                candidates.append(name)

        candidates.append(None)

        return candidates[state]

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(
            repl=self.repl,
            prog='lona_settings',
        )

        argument_parser.add_argument('name', nargs='?')

        arguments = vars(argument_parser.parse_args(argv[1:]))

        server = self.repl.locals['server']

        # write setting
        if arguments['name']:
            self.repl.write(
                pformat(server.settings.get(arguments['name'], None)) + '\n',
            )

            return 0

        # write all settings
        rows = [['Name', 'Value']]

        for key in list(server.settings):
            rows.append(
                [key, pformat(server.settings.get(key))],
            )

        write_table(rows, self.repl.write)



import string
import random

from aiohttp.web import Response

from lona.imports import acquire

REDIRECT_BODY = """
<html>
    <head>
        <meta http-equiv="refresh" content="0; url={url}" />
    </head>
    <body>
    </body>
</html>
""".strip()


def generate_random_string(length):
    choices = string.ascii_letters + string.digits
    return ''.join(random.choice(choices) for _ in range(length))


def generate_session_key(connection):
    return generate_random_string(
        connection.server.settings.SESSIONS_KEY_RANDOM_LENGTH)


class AnonymousUser:
    def __init__(self, session_key=''):
        self.session_key = session_key

    def __repr__(self):
        return f'<AnonymousUser({self.session_key})>'

    def __eq__(self, other):
        if not isinstance(other, AnonymousUser):
            return False

        return self.session_key == other.session_key


class LonaSessionMiddleware:
    async def on_startup(self, data):
        self.generate_session_key = await data.server.run_function_async(
            acquire,
            data.server.settings.SESSIONS_KEY_GENERATOR,
        )

    def handle_connection(self, data):
        http_request = data.connection.http_request
        settings = data.server.settings
        router = data.server.router

        def get_session_key():
            if settings.SESSIONS_KEY_NAME not in http_request.cookies:
                return ''

            return http_request.cookies[settings.SESSIONS_KEY_NAME]

        if not settings.SESSIONS:
            if data.connection.user is None:
                data.connection.user = AnonymousUser()

            return data

        if(not data.connection.interactive and
           not get_session_key()):

            # skip cookie setting and redirecting on non-interactive routes
            # without this exception REST APIs don't work as expected
            match, route, match_info = router.resolve(http_request.path)

            if match and not route.interactive:
                return data

            # set cookie and redirect so the cookie will take effect
            response = Response(
                status=200,
                content_type='text/html',
                text=REDIRECT_BODY.format(url=http_request.path),
            )

            session_key = self.generate_session_key(data.connection)

            response.cookies[settings.SESSIONS_KEY_NAME] = session_key

            return response

        if data.connection.user is None:
            data.connection.user = AnonymousUser(
                session_key=get_session_key(),
            )

        return data



from argparse import RawTextHelpFormatter, ArgumentParser
import logging
import sys
import os

from jinja2 import Environment

from lona.command_line.collect_static import collect_static
from lona.command_line.run_server import run_server
from lona import VERSION_STRING

logger = logging.getLogger('lona')

DESCRIPTION = f"""
Lona {VERSION_STRING}
Documentation: lona-web.org
Code: https://github.com/lona-web-org/lona
""".strip()


def parse_overrides(raw_overrides):
    environment = Environment()
    overrides = {}

    for override in raw_overrides:
        if '=' not in override:
            logger.error(
                "settings overrides: invalid format: '%s'", override)

            continue

        name, value = override.split('=', 1)
        value = environment.compile_expression(value)()

        overrides[name] = value

    return overrides


def handle_command_line(argv):
    parser = ArgumentParser(
        prog='lona',
        formatter_class=RawTextHelpFormatter,
        description=DESCRIPTION,
    )

    # the keyword 'required' is not available in python versions lower than 3.7
    if sys.version_info < (3, 7):
        sub_parsers = parser.add_subparsers(
            dest='command',
        )

    else:
        sub_parsers = parser.add_subparsers(
            dest='command',
            required=True,
        )

    # run-server ##############################################################
    parser_run_server = sub_parsers.add_parser(
        'run-server',
    )

    # logging
    parser_run_server.add_argument(
        '-l',
        '--log-level',
        choices=['debug', 'info', 'warn', 'error', 'critical'],
        default='info',
    )

    parser_run_server.add_argument(
        '--loggers',
        type=str,
        nargs='+',
    )

    parser_run_server.add_argument(
        '--debug-mode',
        choices=['messages', 'views', 'input-events', 'view-events'],
    )

    # settings
    parser_run_server.add_argument(
        '--project-root',
        type=str,
        default=os.getcwd(),
    )

    parser_run_server.add_argument(
        '-s',
        '--settings',
        nargs='+',
        default=[],
    )

    parser_run_server.add_argument(
        '-o',
        '--settings-pre-overrides',
        nargs='+',
        default=[],
    )

    parser_run_server.add_argument(
        '-O',
        '--settings-post-overrides',
        nargs='+',
        default=[],
    )

    # custom
    parser_run_server.add_argument(
        '--host',
        type=str,
        default='localhost',
    )

    parser_run_server.add_argument(
        '--port',
        type=int,
        default=8080,
    )

    parser_run_server.add_argument(
        '--shutdown-timeout',
        type=float,
        default=0.0,
    )

    parser_run_server.add_argument(
        '--shell',
        action='store_true',
    )

    parser_run_server.add_argument(
        '--shell-server-url',
        type=str,
        default='',
    )

    # collect-static ##########################################################
    parser_collect_static = sub_parsers.add_parser(
        'collect-static',
    )

    # logging
    parser_collect_static.add_argument(
        '-l',
        '--log-level',
        choices=['debug', 'info', 'warn', 'error', 'critical'],
        default='warn',
    )

    parser_collect_static.add_argument(
        '--loggers',
        type=str,
        nargs='+',
    )

    parser_collect_static.add_argument(
        '--debug-mode',
        choices=['messages', 'views', 'input-events', 'view-events'],
    )

    # settings
    parser_collect_static.add_argument(
        '--project-root',
        type=str,
        default=os.getcwd(),
    )

    parser_collect_static.add_argument(
        '-s',
        '--settings',
        nargs='+',
        default=[],
    )

    parser_collect_static.add_argument(
        '-o',
        '--settings-pre-overrides',
        nargs='+',
        default=[],
    )

    parser_collect_static.add_argument(
        '-O',
        '--settings-post-overrides',
        nargs='+',
        default=[],
    )

    # custom
    parser_collect_static.add_argument(
        'destination',
    )

    parser_collect_static.add_argument(
        '--clean',
        action='store_true',
        default=False,
    )

    parser_collect_static.add_argument(
        '--silent',
        action='store_true',
        default=False,
    )

    parser_collect_static.add_argument(
        '--dry-run',
        action='store_true',
        default=False,
    )

    # parse argv
    args = parser.parse_args(argv[1:])

    # this can happen on python versions lower than 3.7
    if not args.command:
        exit('no sub command was given')

    args.settings_pre_overrides = parse_overrides(
        args.settings_pre_overrides,
    )

    args.settings_post_overrides = parse_overrides(
        args.settings_post_overrides,
    )

    # run
    if args.command == 'run-server':
        run_server(args)

    elif args.command == 'collect-static':
        collect_static(args)



from rlpython.utils.attribute_table import write_attribute_table
from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table


class LonaRoutesCommand:
    """
    List and test Lona routes
    """

    NAME = 'lona_routes'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(
            repl=self.repl,
            prog='lona_routes',
        )

        argument_parser.add_argument('route-id', nargs='?')
        argument_parser.add_argument('-r', '--resolve', type=str)

        arguments = vars(argument_parser.parse_args(argv[1:]))

        server = self.repl.locals['server']

        # resolve
        if arguments['resolve']:
            path = arguments['resolve']
            match, route, match_info = server.router.resolve(path)

            if not match:
                self.repl.write('No match')

                return

            route_id = server.router.routes.index(route)

            rows = [
                ['Route ID', route_id],
                ['Route', repr(route)],
                ['Match info', repr(match_info)],
            ]

            write_attribute_table(rows, self.repl.write)

            return

        # show given route
        if arguments['route-id']:
            try:
                route_id = int(arguments['route-id'])

                self.repl.write(repr(server.router.routes[route_id]) + '\n')

                return

            except Exception:
                self.repl.write_error('invalid route id\n')

                return 1

        # show all routes
        rows = [['Route ID', 'Route']]

        for index, route in enumerate(server.router.routes):
            rows.append([index, repr(route)])

        write_table(rows, self.repl.write)



class Request:
    def __init__(self, view_runtime, connection):
        self._view_runtime = view_runtime
        self.connection = connection

        self.url = self._view_runtime.url

        if self.url:
            self.GET = dict(self._view_runtime.url.query)
            self.POST = self._view_runtime.post_data or {}

        else:
            self.GET = {}
            self.POST = {}

        self.method = 'POST' if self.POST else 'GET'

    @property
    def interactive(self):
        return self.connection.interactive

    @property
    def user(self):
        return getattr(self.connection, 'user', None)

    @user.setter
    def user(self, new_value):
        self.connection.user = new_value

    @property
    def frontend(self):
        return self._view_runtime.frontend

    @property
    def server(self):
        return self._view_runtime.server

    @property
    def route(self):
        return self._view_runtime.route

    @property
    def match_info(self):
        return self._view_runtime.match_info



from __future__ import annotations

from html.parser import HTMLParser
import logging
import inspect

from lona.html.text_node import TextNode
from lona.html.node import Node

logger = logging.getLogger('lona')

NODE_CLASSES: dict[str, type[Node]] = {}
INPUT_NODE_CLASSES: dict[str, type[Node]] = {}

SELF_CLOSING_TAGS = [
    'area',
    'base',
    'br',
    'col',
    'embed',
    'hr',
    'img',
    'input',
    'link',
    'meta',
    'param',
    'source',
    'track',
    'wbr',
]


def _find_node_classes(module):
    try:
        for attribute_name in dir(module):
            node_class = getattr(module, attribute_name)

            if not inspect.isclass(node_class):
                continue

            if not issubclass(node_class, Node):
                continue

            # input nodes
            if node_class.TAG_NAME == 'input':
                if 'type' not in node_class.ATTRIBUTES:
                    logger.warning(
                        'WARNING: input node %r has no type set',
                        node_class,
                    )

                    continue

                if node_class.ATTRIBUTES['type'] in INPUT_NODE_CLASSES:
                    logger.warning(
                        "WARNING: Two input Node classes with type '%s' were found: %r and %r",
                        node_class.ATTRIBUTES['type'],
                        INPUT_NODE_CLASSES[node_class.ATTRIBUTES['type']],
                        node_class,
                    )

                INPUT_NODE_CLASSES[node_class.ATTRIBUTES['type']] = node_class

            # nodes
            else:
                if node_class.TAG_NAME in NODE_CLASSES:
                    logger.warning(
                        "WARNING: Two Node classes with tag name '%s' were found: %r and %r",
                        node_class.TAG_NAME,
                        NODE_CLASSES[node_class.TAG_NAME],
                        node_class,
                    )

                NODE_CLASSES[node_class.TAG_NAME] = node_class

    except Exception:
        logger.exception('Exception occurred while searching for node classes')


def _setup_node_classes_cache():
    from lona import html

    _find_node_classes(html)


class NodeHTMLParser(HTMLParser):
    CDATA_CONTENT_ELEMENTS = HTMLParser.CDATA_CONTENT_ELEMENTS + ('textarea',)

    def __init__(self, *args, use_high_level_nodes=True, node_classes=None,
                 **kwargs):

        self.use_high_level_nodes = use_high_level_nodes
        self.node_classes = node_classes or {}

        super().__init__(*args, **kwargs)

    def set_current_node(self, node):
        self._node = node

    def get_node_class(self, tag_name, attributes):
        if not self.use_high_level_nodes:
            return Node

        # inputs
        if tag_name == 'input':
            input_type = attributes.get('type', 'text')

            if input_type not in INPUT_NODE_CLASSES:
                return Node

            return INPUT_NODE_CLASSES[input_type]

        # custom node classes
        if tag_name in self.node_classes:
            return self.node_classes[tag_name]

        # nodes from the standard library
        if tag_name in NODE_CLASSES:
            return NODE_CLASSES[tag_name]

        # generic nodes
        return Node

    def handle_startendtag(self, tag, attrs):
        self.handle_starttag(tag, attrs, self_closing=True)

    def handle_starttag(self, tag, attrs, self_closing=False):
        self_closing = self_closing or tag in SELF_CLOSING_TAGS

        # node attributes
        node_attributes = {}

        for name, value in attrs:
            if value is None:
                value = ''

            node_attributes[name] = value

        # tag overrides
        node_kwargs = {}

        node_kwargs['tag_name'] = tag
        node_kwargs['self_closing_tag'] = self_closing

        # setup node
        for key in ('id', 'class', 'style'):
            if key in node_attributes:
                node_kwargs[key] = node_attributes.pop(key)

        node_class = self.get_node_class(tag, node_attributes)
        node = node_class(**node_kwargs)

        # set attributes
        node.attributes.update(node_attributes)

        # setup node
        self._node.append(node)
        if not self_closing:
            self.set_current_node(node)

    def handle_data(self, data):
        text = data

        if self._node.tag_name not in ('pre', 'textarea'):
            text = text.strip()

            if not text:
                return

        # textareas
        if self._node.tag_name == 'textarea':
            self._node.value = data

        # normal nodes
        else:
            self._node.append(TextNode(text))

    def handle_endtag(self, tag):
        if self._node.parent is None:
            raise ValueError(
                f'Invalid html: missing start tag for </{tag}>',
            )
        if tag != self._node.tag_name:
            raise ValueError(
                f'Invalid html: </{self._node.tag_name}> expected, </{tag}> received',
            )

        self.set_current_node(self._node.parent)


def html_string_to_node_list(html_string, use_high_level_nodes=True,
                             node_classes=None):

    root_node = Node()

    html_parser = NodeHTMLParser(
        use_high_level_nodes=use_high_level_nodes,
        node_classes=node_classes or {},
    )

    html_parser.set_current_node(root_node)
    html_parser.feed(html_string)

    if html_parser._node is not root_node:
        raise ValueError(
            f'Invalid html: missing end tag </{html_parser._node.tag_name}>',
        )

    return list(root_node.nodes)



import contextlib


class NotInteractiveError(Exception):
    pass


class Connection:
    def __init__(self, server, http_request, websocket=None):
        self.server = server
        self.http_request = http_request
        self.websocket = websocket

    @property
    def interactive(self):
        return self.websocket is not None

    @property
    def user(self):
        return getattr(self, '_user', None)

    @user.setter
    def user(self, value):
        self._user = value

    def send_str(self, string, wait=True):
        if not self.interactive:
            raise NotInteractiveError

        # this exception gets handled by aiohttp internally and can be ignored
        with contextlib.suppress(ConnectionResetError):
            return self.server.run_coroutine_sync(
                self.websocket.send_str(string),
                wait=wait,
            )

    async def close(self):
        if not self.interactive:
            raise NotInteractiveError

        await self.websocket.close()



from lona.protocol import INPUT_EVENT_TYPE


class EventType:
    def __init__(self, name, symbol):
        self._name = name
        self._symbol = symbol

    def serialize(self):
        return str(self._symbol.value)

    def __repr__(self):
        return self._name

    def __eq__(self, other):
        if not isinstance(other, EventType):
            return False

        return self._symbol == other._symbol


class ChangeEventType:
    def __init__(self, name, symbol, input_delay=0):
        self._name = name
        self._symbol = symbol
        self._input_delay = input_delay

    def __repr__(self):
        if self._input_delay:
            return f'{self._name}(input_delay={self._input_delay})'

        return self._name

    def __call__(self, input_delay=0):
        if not isinstance(input_delay, int):
            raise ValueError('input delay has to be integer')

        return self.__class__(
            name=self._name,
            symbol=self._symbol,
            input_delay=input_delay,
        )

    def serialize(self):
        if self._input_delay:
            return f'{self._symbol.value}:{self._input_delay}'

        return str(self._symbol.value)


CLICK = EventType('CLICK', INPUT_EVENT_TYPE.CLICK)
CHANGE = ChangeEventType('CHANGE', INPUT_EVENT_TYPE.CHANGE)



from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table


class LonaConnectionsCommand:
    """
    List all current connections to Lona
    """

    NAME = 'lona_connections'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(
            repl=self.repl, prog='lona_connections')

        argument_parser.parse_args(argv[1:])

        server = self.repl.locals['server']

        # write connections
        rows = [['User', 'URL']]

        for connection in server.websocket_connections:
            rows.append(
                [repr(connection.user), connection.http_request.url.path],
            )

        write_table(rows, self.repl.write)



from __future__ import annotations

from typing import TypeVar
import importlib
import inspect
import runpy

T = TypeVar('T')


def acquire(import_string, ignore_import_cache=False):
    # scripts
    if '::' in import_string:
        script, attribute_name = import_string.split('::')
        attributes = runpy.run_path(script, run_name=script)

        if attribute_name not in attributes:
            raise ImportError(
                f"script '{script}' has no attribute '{attribute_name}'",
            )

        return attributes[attribute_name]

    # modules
    elif '.' in import_string:
        module_name, attribute_name = import_string.rsplit('.', 1)
        module = importlib.import_module(module_name)

        # ignore import cache
        if ignore_import_cache:
            path = inspect.getfile(module)

            return acquire(f'{path}::{attribute_name}')

        if not hasattr(module, attribute_name):
            raise ImportError(
                f"module '{module_name}' has no attribute '{attribute_name}'",
            )

        return getattr(module, attribute_name)

    else:
        raise TypeError('invalid import string')


def get_file(obj):
    if obj.__module__.endswith('.py'):
        return obj.__module__

    return inspect.getfile(obj)


def get_object_repr(obj):
    return f'<{obj.__class__.__name__} object at {hex(id(obj))}>'


def get_all_subclasses(cls: type[T]) -> set[type[T]]:
    subs = cls.__subclasses__()
    return set(subs).union(s for c in subs for s in get_all_subclasses(c))



from __future__ import annotations

from concurrent.futures import Future
from collections.abc import Awaitable
from typing import overload, TypeVar
from functools import reduce
import contextlib
import operator
import logging
import inspect
import asyncio
import os

from aiohttp.web import WebSocketResponse, FileResponse, HTTPFound, Response
from typing_extensions import Literal
from aiohttp import WSMsgType

from lona.view_runtime_controller import ViewRuntimeController
from lona.middleware_controller import MiddlewareController
from lona.client_pre_compiler import ClientPreCompiler
from lona.static_file_loader import StaticFileLoader
from lona.response_parser import ResponseParser
from lona.templating import TemplatingEngine
from lona.imports import acquire as _acquire
from lona.server_state import ServerState
from lona.shell.shell import embed_shell
from lona.view_loader import ViewLoader
from lona.connection import Connection
from lona.settings import Settings
from lona.protocol import METHOD
from lona.routing import Router
from lona.view import LonaView

DEFAULT_SETTINGS = os.path.join(
    os.path.dirname(__file__),
    'default_settings.py',
)

server_logger = logging.getLogger('lona.server')
http_logger = logging.getLogger('lona.server.http')
websockets_logger = logging.getLogger('lona.server.websockets')

T = TypeVar('T')


class LonaServer:
    def __init__(self, app, project_root, settings_paths=None,
                 settings_pre_overrides=None, settings_post_overrides=None,
                 routes=None):

        self.project_root = os.path.abspath(project_root)

        self.websocket_connections = []
        self._loop = None
        self._worker_pool = None

        server_logger.debug("starting server in '%s'", project_root)

        # setup aiohttp app
        self._app = app
        self._app['lona_server'] = self

        self._app.on_startup.append(self.start)
        self._app.on_shutdown.append(self.stop)

        # setup settings
        server_logger.debug('setup settings')

        self.settings_paths = [
            DEFAULT_SETTINGS,
        ]

        for path in settings_paths or []:
            self.settings_paths.append(
                os.path.normpath(
                    os.path.join(self.project_root, path),
                ),
            )

        self.settings = Settings()

        if settings_pre_overrides:
            server_logger.debug('applying settings pre overrides')

            self.settings.update(settings_pre_overrides)

        for import_string in self.settings_paths:
            server_logger.debug("loading settings from '%s'", import_string)

            self.settings.add(import_string)

        if settings_post_overrides:
            server_logger.debug('applying settings post overrides')

            self.settings.update(settings_post_overrides)

        # setup templating
        server_logger.debug('setup templating')

        self.templating_engine = TemplatingEngine(self)

        # setup server state
        server_logger.debug('setup server state')

        self._state = ServerState(initial_data={})

        # setup routing
        server_logger.debug('setup routing')
        self.router = Router()

        if not routes:
            server_logger.debug("loading routing table from '%s'",
                                self.settings.ROUTING_TABLE)

            routes = self.acquire(self.settings.ROUTING_TABLE)

        if routes:
            self.router.add_routes(*routes)

        else:
            server_logger.warning('routing table is empty')

        # setup middleware controller
        server_logger.debug('setup middleware controller')

        self.middleware_controller = MiddlewareController(self)

        self._app.on_startup.append(
            self.middleware_controller.run_on_startup,
        )

        self._app.on_shutdown.append(
            self.middleware_controller.run_on_shutdown,
        )

        # setup aiohttp routes
        server_logger.debug('setup aiohttp routing')

        static_url = self.settings.STATIC_URL_PREFIX + '{path:.*}'

        server_logger.debug('static url set to %r', static_url)

        self._app.router.add_route(
            '*', static_url, self._handle_static_file_request)

        self._app.router.add_route(
            '*', '/{path_info:.*}', self._handle_http_request)

        # setup view loader
        server_logger.debug('setup view loader')

        self.view_loader = ViewLoader(self)

        # setup response parser
        server_logger.debug('setup response parser')

        self.response_parser = ResponseParser(self)

        # setup views
        server_logger.debug('setup view runtime controller')

        self.view_runtime_controller = ViewRuntimeController(self)

        # setup static files
        self.client_pre_compiler: ClientPreCompiler = ClientPreCompiler(self)

        # the static file loader has to be started last because it does
        # node class discovery which has to happen after all views are imported
        server_logger.debug('setup static file')

        self.static_file_loader = StaticFileLoader(self)

        # finish
        server_logger.debug('setup finish')

    def set_loop(self, loop):
        self._loop = loop

    def set_worker_pool(self, worker_pool):
        self._worker_pool = worker_pool

    @property
    def loop(self):
        return self._loop

    @property
    def worker_pool(self):
        return self._worker_pool

    async def start(self, *args, **kwargs):
        server_logger.debug('start')

        if not self._loop:
            raise RuntimeError('loop is not set')

        if not self._worker_pool:
            raise RuntimeError('worker_pool is not set')

        # run checks
        # tests
        test_names = [
            'TEST_VIEW_START_TIMEOUT',
            'TEST_INPUT_EVENT_TIMEOUT',
        ]

        for test_name in test_names:
            if self.settings.get(test_name, False):
                server_logger.warning('%s is enabled', test_name)

        self.view_runtime_controller.start()

    async def stop(self, *args, **kwargs):
        server_logger.debug('stop')

        await self.run_function_async(self.view_runtime_controller.stop)

        for connection in self.websocket_connections.copy():
            with contextlib.suppress(Exception):
                await connection.websocket.close()

    # asyncio helper ##########################################################
    @overload
    def run_coroutine_sync(
            self,
            coroutine: Awaitable[T],
            wait: None | Literal[True] = True,
    ) -> T:
        ...

    @overload
    def run_coroutine_sync(
            self,
            coroutine: Awaitable[T],
            wait: Literal[False],
    ) -> Future[T]:
        ...

    def run_coroutine_sync(
            self,
            coroutine: Awaitable[T],
            wait: None | bool = True,
    ) -> Future[T] | T:
        future = asyncio.run_coroutine_threadsafe(coroutine, loop=self._loop)

        if wait:
            return future.result()

        return future

    def run_function_async(self, function, *args,
                           executor_name='worker', **kwargs):

        def _function():
            return function(*args, **kwargs)

        return self._loop.run_in_executor(
            self._worker_pool.get_executor(executor_name),
            _function,
        )

    # path helper #############################################################
    def resolve_path(self, path):
        if path.startswith('/'):
            return path

        return os.path.normpath(os.path.join(self.project_root, path))

    def acquire(self, import_string, *args, **kwargs):
        if '::' in import_string:
            script, attribute_name = import_string.split('::')
            script = self.resolve_path(script)

            import_string = f'{script}::{attribute_name}'

        return _acquire(import_string, *args, **kwargs)

    # connection management ###################################################
    async def _setup_connection(self, http_request, websocket=None):
        connection = Connection(self, http_request, websocket)

        handled, data, middleware = \
            await self.middleware_controller.handle_connection(
                connection,
            )

        if websocket is not None:
            self.websocket_connections.append(connection)

        return connection, (handled, data, middleware)

    def _remove_connection(self, connection):
        if connection in self.websocket_connections:
            self.websocket_connections.remove(connection)

    # view helper #############################################################
    def _render_response(self, response_dict):
        if response_dict['file']:
            return FileResponse(response_dict['file'])

        if response_dict['redirect']:
            return HTTPFound(response_dict['redirect'])

        if response_dict['http_redirect']:
            return HTTPFound(response_dict['http_redirect'])

        response = Response(
            status=response_dict['status'],
            content_type=response_dict['content_type'],
            text=response_dict['text'],
        )

        response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'

        return response

    # handle http requests ####################################################
    async def _handle_static_file_request(self, request):
        def _404():
            return Response(
                status=404,
                text='404: Not found',
            )

        if not self.settings.STATIC_FILES_SERVE:
            server_logger.warning(
                'Reverse proxy seems to be misconfigured: a static file request was received but STATIC_FILES_SERVE is disabled',
            )

            return _404()

        rel_path = request.match_info['path']

        abs_path = await self.run_function_async(
            self.static_file_loader.resolve_path,
            rel_path,
            executor_name='static_worker',
        )

        if not abs_path:
            return _404()

        return FileResponse(abs_path)

    async def _handle_websocket_message(self, connection, message):
        websockets_logger.debug(
            '%s message received %s', connection, message.data)

        handled, data, middleware = \
            await self.middleware_controller.handle_websocket_message(
                connection,
                message.data,
            )

        if not handled:
            websockets_logger.debug(
                '%s message got not handled', connection)

    async def _handle_websocket_request(self, http_request):
        websocket = None
        connection = None

        async def close_websocket():
            self.view_runtime_controller.remove_connection(connection)
            self._remove_connection(connection)

            await websocket.close()

            websockets_logger.debug('%s closed', connection)

        # setup websocket
        websocket = WebSocketResponse()
        await websocket.prepare(http_request)

        # setup connection
        connection, middleware_data = await self._setup_connection(
            http_request,
            websocket,
        )

        handled, data, middleware = middleware_data

        # connection got closed by middleware
        if handled:
            if data:
                if not isinstance(data, str):
                    raise RuntimeError(
                        f'{middleware}.handle_connection returned non string data',
                    )

                await connection.send_str(data, wait=False)

            await close_websocket()

            return websocket

        websockets_logger.debug('%s opened', connection)

        # main loop
        try:
            async for message in websocket:
                if message.type == WSMsgType.TEXT:
                    await self._handle_websocket_message(connection, message)

                elif message.type == WSMsgType.PING:
                    await websocket.pong()

                elif message.type in (WSMsgType.CLOSED, WSMsgType.ERROR):
                    break

        except asyncio.CancelledError:
            websockets_logger.debug('CancelledError')

        finally:
            await close_websocket()

        return websocket

    async def _handle_http_request(self, http_request):
        http_logger.debug('http request incoming')

        # resolve path
        http_logger.debug('resolving path %r', http_request.path)

        match, route, match_info = await self.run_function_async(
            self.router.resolve,
            http_request.path,
        )

        if match:
            http_logger.debug('route %s matched', route)

        else:
            http_logger.debug('no route matched')

        # http pass through ###################################################
        if match and route.http_pass_through:
            http_logger.debug('http_pass_through mode')

            # load view
            view = self.view_loader.load(route.view)

            if inspect.isclass(view):
                view = view(
                    server=self,
                    view_runtime=None,
                    request=http_request,
                ).handle_request

            # run view
            if asyncio.iscoroutinefunction(view):
                response = await view(http_request)

            else:
                response = await self.run_function_async(
                    view,
                    http_request,
                    executor_name='runtime_worker',
                )

            if asyncio.iscoroutine(response):
                response = await response

            # render and return response
            if isinstance(response, dict):
                return await self.run_function_async(
                    self._render_response,
                    response,
                )

            return response

        # websocket requests ##################################################
        if(http_request.method == 'GET' and
           http_request.headers.get('upgrade', '').lower() == 'websocket'):

            return await self._handle_websocket_request(http_request)

        # non interactive view or frontend ####################################
        # setup connection
        try:
            connection, middleware_data = await self._setup_connection(
                http_request,
            )

        except Exception:
            http_logger.exception(
                'Exception occurred while setting connection up',
            )

            return Response(status=500, body='500: Internal Error')

        handled, data, middleware = middleware_data

        # connection got closed by middleware
        if handled:
            if data is not None:
                if isinstance(data, dict):
                    return self._render_response(data)

                return data

            return Response(status=503, body='503: Service Unavailable')

        post_data = await http_request.post()

        response_dict = await self.run_function_async(
            self.view_runtime_controller.handle_view_message,
            connection=connection,
            window_id=None,
            view_runtime_id=None,
            method=METHOD.VIEW,
            payload=[str(http_request.url), post_data],
            executor_name='runtime_worker',
        )

        return self._render_response(response_dict)

    # state ###################################################################
    @property
    def state(self):
        return self._state

    # helper ##################################################################
    def embed_shell(self, _locals=None):
        if _locals is None:
            _locals = {}
        _locals['server'] = self

        return embed_shell(self, locals=_locals)

    def get_running_views_count(self, *args, **kwargs):
        return self.view_runtime_controller.get_running_views_count(
            *args,
            **kwargs,
        )

    def view_is_already_running(self, request):
        view_runtime = self.view_runtime_controller.get_running_view_runtime(
            user=request.user,
            route=request.route,
            match_info=request.match_info,
        )

        return bool(view_runtime)

    def get_connection_count(self, user):
        count = 0

        for connection in self.websocket_connections.copy():
            if connection.user == user:
                count += 1

        return count

    def get_connected_user_count(self):
        user = []

        def add_user(new_user):
            for _user in user:
                if new_user == _user:
                    return

            user.append(new_user)

        for connection in self.websocket_connections.copy():
            add_user(connection.user)

        return len(user)

    def get_template(self, *args, **kwargs):
        return self.templating_engine.get_template(*args, **kwargs)

    def render_string(self, *args, **kwargs):
        return self.templating_engine.render_string(*args, **kwargs)

    def render_template(self, *args, **kwargs):
        return self.templating_engine.render_template(*args, **kwargs)

    def get_view_class(self, route=None, import_string=None, url=None):
        args = [bool(route), bool(import_string), bool(url)]

        if reduce(operator.xor, args, True):
            raise ValueError('too many or too few arguments given')

        if route:
            return self.view_loader.load(route.view)

        if import_string:
            return self.view_loader.load(import_string)

        if url:
            success, route, match_info = self.router.resolve(url)

            if not success:
                return None

            return self.view_loader.load(route.view)

    def reverse(self, *args, **kwargs):
        return self.router.reverse(*args, **kwargs)

    def fire_view_event(
            self,
            name: str,
            data: dict | None = None,
            view_classes: type[LonaView] | list[type[LonaView]] | None = None,
    ) -> None:

        self.view_runtime_controller.fire_view_event(
            name=name,
            data=data,
            view_classes=view_classes,
        )



from lona.events.event_types import CLICK
from lona.html.node import Node


# simple html nodes ###########################################################
class Nav(Node):
    TAG_NAME = 'nav'


class Main(Node):
    TAG_NAME = 'main'


class Section(Node):
    TAG_NAME = 'section'


class Script(Node):
    TAG_NAME = 'script'


class NoScript(Node):
    TAG_NAME = 'noscript'


class H1(Node):
    TAG_NAME = 'h1'


class H2(Node):
    TAG_NAME = 'h2'


class H3(Node):
    TAG_NAME = 'h3'


class H4(Node):
    TAG_NAME = 'h4'


class H5(Node):
    TAG_NAME = 'h5'


class H6(Node):
    TAG_NAME = 'h6'


class Div(Node):
    TAG_NAME = 'div'


class Span(Node):
    TAG_NAME = 'span'


class P(Node):
    TAG_NAME = 'p'


class I(Node):  # NOQA: E742
    TAG_NAME = 'i'


class Strong(Node):
    TAG_NAME = 'strong'


class Small(Node):
    TAG_NAME = 'small'


class Ul(Node):
    TAG_NAME = 'ul'


class Ol(Node):
    TAG_NAME = 'ol'


class Li(Node):
    TAG_NAME = 'li'


class Table(Node):
    TAG_NAME = 'table'


class THead(Node):
    TAG_NAME = 'thead'


class TBody(Node):
    TAG_NAME = 'tbody'


class TFoot(Node):
    TAG_NAME = 'tfoot'


class ColGroup(Node):
    TAG_NAME = 'colgroup'


class Tr(Node):
    TAG_NAME = 'tr'


class Th(Node):
    TAG_NAME = 'th'


class Td(Node):
    TAG_NAME = 'td'


class Canvas(Node):
    TAG_NAME = 'canvas'


class Hr(Node):
    TAG_NAME = 'hr'
    SELF_CLOSING_TAG = True


class Br(Node):
    TAG_NAME = 'br'
    SELF_CLOSING_TAG = True


# inputs ######################################################################
class Form(Node):
    TAG_NAME = 'form'


class Fieldset(Node):
    TAG_NAME = 'fieldset'


class Label(Node):
    TAG_NAME = 'label'


class Button(Node):
    TAG_NAME = 'button'
    EVENTS = [CLICK]

    def __init__(self, *args, disabled=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.disabled = disabled

    @property
    def disabled(self):
        return 'disabled' in self.attributes

    @disabled.setter
    def disabled(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('disabled is a boolean property')

        if new_value:
            self.attributes['disabled'] = ''
        else:
            del self.attributes['disabled']


class Submit(Node):
    TAG_NAME = 'input'
    SELF_CLOSING_TAG = True

    ATTRIBUTES = {
        'type': 'submit',
        'value': 'Submit',
    }


# complex html nodes ##########################################################
class Img(Node):
    TAG_NAME = 'img'
    SELF_CLOSING_TAG = True


class A(Node):
    TAG_NAME = 'a'

    def __init__(self, *args, interactive=None, **kwargs):
        super().__init__(*args, **kwargs)

        # This is necessary because 'interactive' uses 'ignore' internally.
        # If interactive were set to True by default ignore=True wouldn't
        # work as expected.
        if interactive is not None:
            self.interactive = interactive

    @property
    def interactive(self):
        return not self.ignore

    @interactive.setter
    def interactive(self, value):
        self.ignore = not value

    def set_href(self, href):
        self.attributes['href'] = str(href)

    def get_href(self, href):
        return self.attributes['href']


class Pre(Node):
    TAG_NAME = 'pre'

    def write(self, string):
        string = str(string)

        if len(self.nodes) > 0 and isinstance(self.nodes[-1], str):
            self.nodes[-1] = self.nodes[-1] + string

        else:
            self.nodes.append(string)

    def write_line(self, string):
        string = str(string) + '\n'

        self.nodes.append(string)



from lona.protocol import INPUT_EVENT_TYPE
from lona.html import Widget


class InputEvent:
    def __init__(self, request, payload, document, connection, window_id):
        self.request = request
        self.payload = payload
        self.document = document
        self.connection = connection
        self.window_id = window_id

        self.data = {}
        self.nodes = []
        self.node = None
        self.tag_name = ''
        self.id_list = []
        self.class_list = []

        self.event_id = payload[0]

        # parse input event type
        if isinstance(payload[1], str):
            self.type = INPUT_EVENT_TYPE.CUSTOM
            self.name = payload[1]
            self.data = payload[2]
            self.node_info = payload[3:]

        elif payload[1] == INPUT_EVENT_TYPE.CLICK:
            self.type = INPUT_EVENT_TYPE.CLICK
            self.name = 'click'
            self.data = payload[2]
            self.node_info = payload[3:]

        elif payload[1] == INPUT_EVENT_TYPE.CHANGE:
            self.type = INPUT_EVENT_TYPE.CHANGE
            self.name = 'change'
            self.data = payload[2]
            self.node_info = payload[3:]

        # find node
        # node info contains a lona node id
        if self.node_info[0]:
            self.nodes = document.get_node(node_id=self.node_info[0])

            if self.nodes:
                self.node = self.nodes[0]

        self.tag_name = self.node_info[1]
        self.id_list = (self.node_info[2] or '').split(' ')
        self.class_list = (self.node_info[3] or '').split(' ')

    def node_has_id(self, name):
        if self.node is None:
            return name in self.id_list

        if isinstance(self.node, Widget):
            if not hasattr(self.node, 'has_id'):
                return False

        return self.node.has_id(name)

    def node_has_class(self, name):
        if self.node is None:
            return name in self.class_list

        if isinstance(self.node, Widget):
            if not hasattr(self.node, 'has_class'):
                return False

        return self.node.has_class(name)



from __future__ import annotations

from collections.abc import Iterable
from textwrap import indent

from lona.html.attribute_dict import AttributeDict, StyleDict
from lona.html.attribute_list import ClassList, IDList
from lona.html.node_event_list import NodeEventList
from lona.events import ChangeEventType, EventType
from lona.html.abstract_node import AbstractNode
from lona.html.node_list import NodeList
from lona.protocol import NODE_TYPE


def parse_style_string(style_string: str) -> dict[str, str]:
    values = {}

    for css_rule in style_string.split(';'):
        if not css_rule.strip():
            continue

        if ':' not in css_rule:
            raise ValueError(f'Invalid style string: {style_string}')

        name, value = css_rule.split(':', 1)

        values[name.strip()] = value.strip()

    return values


class Node(AbstractNode):
    TAG_NAME = 'html'
    SELF_CLOSING_TAG = False
    ID_LIST: list[str] = []
    CLASS_LIST: list[str] = []
    STYLE: dict[str, str] | str = {}
    ATTRIBUTES: dict[str, str] = {}
    EVENTS: list[EventType | ChangeEventType] = []

    def __init__(self, *args, tag_name=None, self_closing_tag=None, **kwargs):
        self._id_list = IDList(self, self.ID_LIST)
        self._class_list = ClassList(self, self.CLASS_LIST)
        self._style = StyleDict(self, self.STYLE)
        self._attributes = AttributeDict(self, self.ATTRIBUTES)
        self._nodes = NodeList(self)
        self._events = NodeEventList(self, self.EVENTS)

        # tag overrides
        self.tag_name = tag_name or self.TAG_NAME

        if self_closing_tag is None:
            self.self_closing_tag = self.SELF_CLOSING_TAG

        else:
            self.self_closing_tag = self_closing_tag

        # args (nodes)
        for arg in args:
            if isinstance(arg, (AbstractNode, str)):
                self.append(arg)

            elif isinstance(arg, Iterable):
                for node in list(arg):
                    self.append(node)

            else:
                self.append(arg)

        # kwargs (attributes)
        for name, value in kwargs.items():

            # remove underscores from attributes
            # this makes kwargs like '_class' possible to prevent clashes
            # with python grammar
            if '_' in name:
                name = name.replace('_', '-')

                if name.startswith('-'):
                    name = name[1:]

            # lona classes
            if name == 'ignore':
                setattr(self, name, value)

            # patchable attributes
            elif name == 'id':
                if not isinstance(value, (str, list)):
                    raise ValueError(
                        'id has to be string or list of strings',
                    )

                if isinstance(value, str):
                    value = value.split(' ')

                self._id_list.extend(value)

            elif name == 'class':
                if not isinstance(value, (str, list)):
                    raise ValueError(
                        'class has to be string or list of strings',
                    )

                if isinstance(value, str):
                    value = value.split(' ')

                self._class_list.extend(value)

            elif name == 'style':
                if not isinstance(value, (dict, str)):
                    raise ValueError('style has to be dict or string')

                if isinstance(value, str):
                    value = parse_style_string(value)

                self._style.update(value)

            elif name == 'attributes':
                if not isinstance(value, dict):
                    raise ValueError('attributes has to be dict')

                self._attributes.update(value)

            elif name == 'events':
                if not isinstance(value, list):
                    raise ValueError('events have to be list')

                self._events.extend(value)

            elif name == 'nodes':
                if not isinstance(value, list):
                    value = [value]

                self._nodes._reset(value)

            elif name == 'handle-change':  # '_' was replaced to '-' above
                if not callable(value):
                    raise TypeError('handle_change has to be a function')

                self.handle_change = value

            elif name == 'handle-click':  # '_' was replaced to '-' above
                if not callable(value):
                    raise TypeError('handle_click has to be a function')

                self.handle_click = value

            # misc attributes
            else:
                self._attributes[name] = value

    # node attributes  ########################################################
    # id_list
    @property
    def id_list(self):
        return self._id_list

    @id_list.setter
    def id_list(self, value):
        self._id_list._reset(value)

    # class_list
    @property
    def class_list(self):
        return self._class_list

    @class_list.setter
    def class_list(self, value):
        self._class_list._reset(value)

    # style
    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, value):
        self._style._reset(value)

    # attributes
    @property
    def attributes(self):
        return self._attributes

    @attributes.setter
    def attributes(self, value):
        self._attributes._reset(value)

    # events
    @property
    def events(self):
        return self._events

    @events.setter
    def events(self, value):
        self._events._reset(value)

    # nodes
    @property
    def nodes(self):
        return self._nodes

    @nodes.setter
    def nodes(self, value):
        self._nodes._reset(value)

    # lona attribute helper ###################################################
    def has_class(self, class_name):
        return class_name in self._class_list

    def has_id(self, id_name):
        return id_name in self._id_list

    # lona attributes #########################################################
    @property
    def ignore(self):
        return 'data-lona-ignore' in self._attributes

    @ignore.setter
    def ignore(self, value):
        if not isinstance(value, bool):
            raise TypeError('ignore is a boolean property')

        if value:
            self._attributes['data-lona-ignore'] = ''

        else:
            del self._attributes['data-lona-ignore']

    # serialization ###########################################################
    def _serialize(self):
        return [
            NODE_TYPE.NODE,
            self.id,
            self.tag_name,
            self._id_list._serialize(),
            self._class_list._serialize(),
            self._style._serialize(),
            self._attributes._serialize(),
            self._nodes._serialize(),
        ]

    # node list helper ########################################################
    def insert(self, index, node):
        self._nodes.insert(index, node)

    def append(self, node):
        self._nodes.append(node)

    def remove(self, node=None):
        if not node:
            if not self._parent:
                raise RuntimeError('node has no parent node')

            self._parent.remove(self)

            return

        self._nodes.remove(node)

    def pop(self, index):
        return self._nodes.pop(index)

    def clear(self):
        self._nodes.clear()

    def __getitem__(self, index):
        return self._nodes[index]

    def __setitem__(self, index, value):
        self._nodes[index] = value

    def __iter__(self):
        return self._nodes.__iter__()

    def __len__(self):
        return self._nodes.__len__()

    def __bool__(self):
        return True

    # string representation ###################################################
    def __str__(self, node_string=None, skip_value=False):
        with self.lock:
            # opening tag
            string = f'<{self.tag_name} data-lona-node-id="{self.id}"'

            if self.id_list:
                string += f' id="{self.id_list}"'

            if self.class_list:
                string += f' class="{self.class_list}"'

            if self.style:
                string += f' style="{self.style.to_sub_attribute_string()}"'

            if self.attributes:
                string += ' '

                string += self.attributes.to_attribute_string(
                    skip_value=skip_value,
                )

            if self.self_closing_tag:
                string += ' />'

            else:
                string += '>'

            # nodes
            if node_string:
                string += '\n'
                string += indent(node_string, '  ')
                string += '\n'

            elif self.nodes:
                string += '\n'
                string += indent(str(self.nodes), '  ')
                string += '\n'

            # closing tag
            if not self.self_closing_tag:
                string += f'</{self.tag_name}>'

            return string

    def __repr__(self):
        return self.__str__()

    # HTML helper #############################################################
    def set_text(self, text):
        self.nodes = [str(text)]

    def get_text(self):
        with self.lock:
            text = []

            for node in self.nodes:
                text.append(node.get_text())

            text = ' '.join(text)

            # strip starting and ending whitespaces but preserve \n
            # this is important for nodes like <pre> which are newline aware
            while text.startswith('\t') or text.startswith(' '):
                text = text[1:]

            while text.endswith('\t') or text.endswith(' '):
                text = text[:1]

            return text

    def hide(self):
        self.style['display'] = 'none'

    def show(self):
        with self.lock:
            if 'display' in self.style and self.style['display'] == 'none':
                del self.style['display']



from __future__ import annotations

import contextlib
import logging

from yarl import URL

from lona.protocol import encode_http_redirect, METHOD
from lona.events.view_event import ViewEvent
from lona.view_runtime import ViewRuntime
from lona.exceptions import ServerStop
from lona.view import LonaView

input_events_logger = logging.getLogger('lona.input_events')
view_events_logger = logging.getLogger('lona.view_events')
views_logger = logging.getLogger('lona.views')


class ViewRuntimeController:
    def __init__(self, server):
        self.server = server

        self._view_runtimes = {
            # contains: view_runtime_id: view_runtime
        }

    def start(self):
        pass

    def stop(self):
        for view_runtime in self.iter_view_runtimes():
            view_runtime.stop(reason=ServerStop)

    # helper ##################################################################
    def get_view_runtime(self, view_runtime_id):
        return self._view_runtimes.get(view_runtime_id, None)

    def get_running_view_runtime(self, user, route, match_info):
        for view_runtime in self.iter_view_runtimes():
            if(view_runtime.start_connection.user == user and
               view_runtime.is_daemon and
               view_runtime.route == route and
               view_runtime.match_info == match_info):

                return view_runtime

    def remove_view_runtime(self, view_runtime):
        with contextlib.suppress(KeyError):
            self._view_runtimes.pop(view_runtime.view_runtime_id)

        view_runtime.run_cleanup_hook()

    def iter_view_runtimes(self):
        view_runtime_ids = list(self._view_runtimes.keys())

        for view_runtime_id in view_runtime_ids:
            view_runtime = self.get_view_runtime(view_runtime_id)

            if not view_runtime:
                continue

            yield view_runtime

    def remove_connection(self, connection, window_id=None):
        for view_runtime in self.iter_view_runtimes():
            view_runtime.remove_connection(
                connection=connection,
                window_id=window_id,
            )

    def get_running_views_count(self, user):
        count = 0

        for view_runtime in self.iter_view_runtimes():
            if(view_runtime.start_connection and
               view_runtime.start_connection.user == user):

                count += 1

        return count

    # lona messages ###########################################################
    def handle_input_event_message(self, connection, window_id,
                                   view_runtime_id, method, payload):

        view_runtime = self.get_view_runtime(view_runtime_id)

        if not view_runtime:
            input_events_logger.debug(
                'event #%s: runtime id is unknown. event is skipped',
                payload[0],
            )

            return

        if connection.user != view_runtime.start_connection.user:
            input_events_logger.debug(
                'event #%s: connection.user is not authorized. event is skipped',
                payload[0],
            )

            return

        input_events_logger.debug(
            'event #%s: gets handled by runtime #%s',
            payload[0],
            view_runtime_id,
        )

        view_runtime.handle_input_event(
            connection=connection,
            window_id=window_id,
            payload=payload,
        )

    def handle_client_error_message(self, connection, window_id,
                                    view_runtime_id, method, payload):

        view_runtime = self.get_view_runtime(view_runtime_id)

        if not view_runtime:
            views_logger.debug(
                'runtime id is unknown. event is skipped',
                payload[0],
            )

            return

        if connection.user != view_runtime.start_connection.user:
            views_logger.debug(
                'connection.user is not authorized. client error is skipped',
            )

            return

        views_logger.debug(
            'client error gets handled by runtime #%s',
            view_runtime_id,
        )

        view_runtime.handle_client_error(
            connection=connection,
            window_id=window_id,
            payload=payload,
        )

    def handle_view_message(self, connection, window_id, view_runtime_id,
                            method, payload):

        url, post_data = payload

        # resolve url
        url_object = URL(url)
        match, route, match_info = self.server.router.resolve(url_object.path)

        # route is not interactive; issue a http redirect
        if(connection.interactive and
           match and
           (route.http_pass_through or not route.interactive)):

            views_logger.debug(
                'route is not interactive; issue a http redirect',
            )

            message = encode_http_redirect(
                window_id=window_id,
                view_runtime_id=None,
                target_url=url,
            )

            connection.send_str(message)

            views_logger.debug('message handled')

            return

        # reconnect to daemonized view runtime ################################
        if connection.interactive:
            views_logger.debug('removing old connections')

            self.remove_connection(
                connection,
                window_id=window_id,
            )

            # search for a running view runtime to reconnect to
            views_logger.debug('searching for running view runtime')

            running_view_runtime = self.get_running_view_runtime(
                user=connection.user,
                route=route,
                match_info=match_info,
            )

            if running_view_runtime and not running_view_runtime.is_stopped:
                views_logger.debug('reconnecting to %r', running_view_runtime)

                running_view_runtime.reconnect_connection(
                    connection=connection,
                    window_id=window_id,
                    url=url,
                )

                views_logger.debug('message handled')

                return

        # start new view runtime ##############################################
        # remove previous running runtime
        if connection.interactive and running_view_runtime:
            views_logger.debug('removing previous runtime')

            self.server.view_runtime_controller.remove_view_runtime(
                view_runtime=running_view_runtime,
            )

        # start nev runtime
        views_logger.debug('trying to start a new view runtime')

        frontend = False

        if not connection.interactive and match and route.interactive:
            frontend = True

            views_logger.debug('running in frontend mode')

        view_runtime = ViewRuntime(
            server=self.server,
            url=url,
            route=route,
            match_info=match_info,
            post_data=post_data or {},
            start_connection=connection,
            frontend=frontend,
        )

        if connection.interactive:
            view_runtime.add_connection(
                connection=connection,
                window_id=window_id,
                url=url,
            )

        # run middlewares
        response_dict = view_runtime.run_middlewares(
            connection=connection,
            window_id=window_id,
            url=url,
        )

        # request got handled by a middleware
        if response_dict:
            views_logger.debug('connection was refused by middleware')

            return response_dict

        self._view_runtimes[view_runtime.view_runtime_id] = view_runtime

        # start view runtime
        views_logger.debug('starting new view runtime')

        if connection.interactive:
            # If the connection is interactive this method got called
            # by the LonaMessageMiddleware and therefore run in the generic
            # worker pool up this point.
            # To release the worker we have to reschedule to the
            # 'runtime_worker' pool.

            self.server.run_function_async(
                view_runtime.start,
                executor_name='runtime_worker',
            )

        else:
            return view_runtime.start()

        views_logger.debug('message handled')

    def handle_lona_message(self, connection, window_id, view_runtime_id,
                            method, payload):

        """
        this method gets called by the
        lona.middlewares.LonaMessageMiddleware.process_websocket_message

        """

        # views
        if method == METHOD.VIEW:
            views_logger.debug('view message decoded: %r', payload)

            self.handle_view_message(
                connection=connection,
                window_id=window_id,
                view_runtime_id=view_runtime_id,
                method=method,
                payload=payload,
            )

        # input events
        elif method == METHOD.INPUT_EVENT:
            input_events_logger.debug('event #%s: decoded', payload[0])

            self.handle_input_event_message(
                connection=connection,
                window_id=window_id,
                view_runtime_id=view_runtime_id,
                method=method,
                payload=payload,
            )

        # client error
        elif method == METHOD.CLIENT_ERROR:
            views_logger.debug('client error decoded')

            self.handle_client_error_message(
                connection=connection,
                window_id=window_id,
                view_runtime_id=view_runtime_id,
                method=method,
                payload=payload,
            )

    # events ##################################################################
    def run_view_event_hook(
            self,
            view_runtime: ViewRuntime,
            view_event: ViewEvent,
    ) -> None:

        try:
            view_events_logger.debug(
                'running %r with %r',
                view_runtime.view.on_view_event,
                view_event,
            )

            return_value = view_runtime.view.on_view_event(view_event)

        except Exception:
            view_events_logger.exception(
                'Exception raised while running %r',
                view_runtime.view.on_view_event,
            )

            return

        if not isinstance(return_value, (dict, type(None))):
            exception = ValueError(f'{repr(view_runtime.view.on_view_event)} returned an unexpected type ({repr(return_value)})')

            view_runtime.issue_500_error(exception)

        if isinstance(return_value, dict):
            response_parser = self.server.response_parser

            try:
                response_dict = response_parser.parse_event_response_dict(
                    return_value,
                )

                view_runtime.handle_raw_response_dict(response_dict)

            except Exception as exception:
                view_runtime.issue_500_error(exception)

            return

        # send html updates if necessary
        if view_runtime.document.is_dirty:
            view_runtime.view.show()

    def fire_view_event(
            self,
            name: str,
            data: dict | None = None,
            view_classes: type[LonaView] | list[type[LonaView]] | None = None,
    ) -> None:

        data = data or {}
        view_classes = view_classes or []

        if not isinstance(view_classes, list):
            view_classes = [view_classes]

        view_events_logger.debug(
            'sending ViewEvent(name=%r, data=%r, view_classes=%r)',
            name,
            data,
            view_classes,
        )

        for view_runtime in self.iter_view_runtimes():
            if view_classes and view_runtime.view_class not in view_classes:
                continue

            # if on_view_event is not defined by the view class
            # it is unnecessary to start a thread
            if view_runtime.view_class.on_view_event == LonaView.on_view_event:
                view_events_logger.debug(
                    '%r skipped (view class does not implement on_view_event())',
                    view_runtime.view,
                )

                continue

            view_event = ViewEvent(
                name=name,
                data=data.copy(),
                view_classes=view_classes.copy(),
            )

            self.server.run_function_async(
                self.run_view_event_hook,
                view_runtime,
                view_event,
            )



from lona.html.parsing import html_string_to_node_list
from lona.html.text_node import TextNode
from lona.html.widget import Widget


class HTML(Widget):
    def __init__(self, *nodes, use_high_level_nodes=True, node_classes=None):
        self.nodes = []

        for node in nodes:

            # strings
            if isinstance(node, str):

                # escaped text
                if node.startswith('\\'):
                    self.nodes.append(TextNode(node[1:]))

                # html string
                elif '<' in node or '>' in node:
                    if len(nodes) > 1:
                        self.nodes.append(HTML(node))

                    else:
                        self.nodes = html_string_to_node_list(
                            html_string=node,
                            use_high_level_nodes=use_high_level_nodes,
                            node_classes=node_classes or {},
                        )

                else:
                    self.nodes.append(TextNode(node))

            # lona nodes
            else:
                self.nodes.append(node)

    def insert(self, *args, **kwargs):
        return self.nodes.insert(*args, **kwargs)

    def append(self, *args, **kwargs):
        return self.nodes.append(*args, **kwargs)

    def remove(self, *args, **kwargs):
        return self.nodes.remove(*args, **kwargs)

    def clear(self, *args, **kwargs):
        return self.nodes.clear(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        return self.nodes.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        return self.nodes.__setitem__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        return self.nodes.__iter__(*args, **kwargs)

    def __contains__(self, *args, **kwargs):
        return self.nodes.__contains__(*args, **kwargs)

    def __bool__(self, *args, **kwargs):
        return self.nodes.__bool__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self.nodes.__len__(*args, **kwargs)



try:
    from .exceptions import *  # NOQA: F403
    from .routing import MATCH_ALL, Route
    from .errors import *  # NOQA: F403
    from .view import LonaView
    from .app import LonaApp

except ImportError:
    # this happens while packaging and can be ignored

    pass

VERSION = (1, 7, 4)
VERSION_STRING = '.'.join(str(i) for i in VERSION)



from __future__ import annotations

from typing import Any
import contextlib

from lona.events.event_types import CHANGE
from lona.html.node import Node

DEFAULT_INPUT_DELAY = 300


class TextInput(Node):
    TAG_NAME = 'input'
    SELF_CLOSING_TAG = True
    INPUT_ATTRIBUTE_NAME = 'value'

    ATTRIBUTES = {
        'type': 'text',
    }

    def __init__(self, value='', disabled=False, readonly=False,
                 bubble_up=False, input_delay=DEFAULT_INPUT_DELAY, **kwargs):
        super().__init__(**kwargs)
        self.value = value
        self.disabled = disabled
        self.readonly = readonly
        self.bubble_up = bubble_up
        self.events.add(CHANGE(input_delay))

    def handle_input_event(self, input_event):
        # Data binding nodes catch their own change events and synchronize
        # their internal value. When setting their value, a HTML patch,
        # containing the set value, gets created that gets distributed to all
        # connections that are connected to the runtime. By default the input
        # timeout is set to 300ms. When the user # is typing fast, it can
        # happen that a patch gets applied when the user is still typing,
        # resetting the input to a previous state. For the
        # user it looks like the input looses characters while typing.

        # The solution for this problem is to don't send patches back to users
        # who issued them.

        if input_event.name == 'change':
            self.attributes.__setitem__(
                self.INPUT_ATTRIBUTE_NAME,
                input_event.data,
                issuer=(input_event.connection, input_event.window_id),
            )

            input_event = self.handle_change(input_event)

        elif input_event.name == 'click':
            input_event = self.handle_click(input_event)

        else:
            return input_event

        if self.bubble_up:
            return input_event

    # properties ##############################################################
    # value
    @property
    def value(self):
        return self.attributes.get(self.INPUT_ATTRIBUTE_NAME, '')

    @value.setter
    def value(self, new_value):
        self.attributes[self.INPUT_ATTRIBUTE_NAME] = new_value

    # disabled
    @property
    def disabled(self):
        return 'disabled' in self.attributes

    @disabled.setter
    def disabled(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('disabled is a boolean property')

        if new_value:
            self.attributes['disabled'] = ''

        else:
            del self.attributes['disabled']

    # readonly
    @property
    def readonly(self):
        return 'readonly' in self.attributes

    @readonly.setter
    def readonly(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('readonly is a boolean property')

        if new_value:
            self.attributes['readonly'] = ''

        else:
            del self.attributes['readonly']


class TextArea(TextInput):
    TAG_NAME = 'textarea'
    SELF_CLOSING_TAG = False
    ATTRIBUTES: dict[str, str] = {}

    def __repr__(self):
        return self.__str__(
            skip_value=True,
            node_string=self.attributes.get('value', ''),
        )


class CheckBox(TextInput):
    INPUT_ATTRIBUTE_NAME = 'checked'
    ATTRIBUTES = {
        'type': 'checkbox',
    }

    def __init__(self, value=False, disabled=False, readonly=False,
                 bubble_up=False, **kwargs):

        super().__init__(value, disabled, readonly, bubble_up, **kwargs)

    @property
    def value(self) -> bool:
        value = self.attributes.get(self.INPUT_ATTRIBUTE_NAME, False)

        if value == '':  # is possible after parsing HTML string
            return True

        return bool(value)

    @value.setter
    def value(self, new_value: bool) -> None:
        if not isinstance(new_value, bool):
            raise TypeError('value is a boolean property')

        # Don't need to remove `checked` attribute if False because
        # it is a special property and is handled differently by js client
        # (search for `property_names` in /lona/client)
        self.attributes[self.INPUT_ATTRIBUTE_NAME] = new_value


class NumberInput(TextInput):
    ATTRIBUTES = {
        'type': 'number',
    }

    def __init__(
            self,
            value: None | float = None,
            min: None | float = None,
            max: None | float = None,
            step: None | float = None,
            disabled: bool = False,
            readonly: bool = False,
            bubble_up: bool = False,
            input_delay: int = DEFAULT_INPUT_DELAY,
            **kwargs: Any,
    ) -> None:
        super().__init__(
            value,
            disabled,
            readonly,
            bubble_up,
            input_delay,
            **kwargs,
        )
        self.min = min
        self.max = max
        self.step = step

    # properties ##############################################################
    # value
    @property
    def value(self) -> None | float:
        value = self.attributes.get(self.INPUT_ATTRIBUTE_NAME, '')
        if value != '':
            with contextlib.suppress(ValueError):
                return float(value)
        return None

    @value.setter
    def value(self, new_value: None | float) -> None:
        if new_value is not None and not isinstance(new_value, (int, float)):
            raise TypeError('value should be None, int or float')

        if new_value is None:
            self.attributes[self.INPUT_ATTRIBUTE_NAME] = ''
        else:
            self.attributes[self.INPUT_ATTRIBUTE_NAME] = new_value

    # min
    @property
    def min(self) -> None | float:
        value = self.attributes.get('min', None)
        if value is not None:
            with contextlib.suppress(ValueError):
                return float(value)
        return None

    @min.setter
    def min(self, new_value: None | float) -> None:
        if new_value is not None and not isinstance(new_value, (int, float)):
            raise TypeError('min should be None, int or float')

        if new_value is None:
            del self.attributes['min']
        else:
            self.attributes['min'] = new_value

    # max
    @property
    def max(self) -> None | float:
        value = self.attributes.get('max', None)
        if value is not None:
            with contextlib.suppress(ValueError):
                return float(value)
        return None

    @max.setter
    def max(self, new_value: None | float) -> None:
        if new_value is not None and not isinstance(new_value, (int, float)):
            raise TypeError('max should be None, int or float')

        if new_value is None:
            del self.attributes['max']
        else:
            self.attributes['max'] = new_value

    # step
    @property
    def step(self) -> None | float:
        value = self.attributes.get('step', None)
        if value is not None:
            with contextlib.suppress(ValueError):
                return float(value)
        return None

    @step.setter
    def step(self, new_value: None | float) -> None:
        if new_value is not None and not isinstance(new_value, (int, float)):
            raise TypeError('step should be None, int or float')

        if new_value is None:
            del self.attributes['step']
        else:
            self.attributes['step'] = new_value



import logging

from lona.html.abstract_node import AbstractNode
from lona._json import dumps

logger = logging.getLogger('lona.response_parser')


class ResponseParser:
    def __init__(self, server):
        self.server = server

    def parse_event_response_dict(
            self,
            raw_response_dict: dict,
    ) -> dict:

        """
        Parse response dicts coming from event handlers.

        response dicts can only be redirects or http redirects.

        examples:
            {'redirect': '/foo'}
            {'http_redirect': '/bar'}
        """

        supported_keys = ('redirect', 'http_redirect')

        if(len(raw_response_dict.keys()) > 1 or
           list(raw_response_dict.keys())[0] not in supported_keys):

            raise ValueError(f'response dict has unexpected keys ({repr(list(raw_response_dict.keys()))})')

        redirect_type, redirect_target = list(raw_response_dict.items())[0]

        if not isinstance(redirect_target, str):
            raise ValueError('redirect target has to be string')

        return raw_response_dict

    def render_response_dict(self, raw_response_dict, view_name):
        response_dict = {
            'status': 200,
            'content_type': 'text/html',
            'text': '',
            'file': '',
            'redirect': '',
            'http_redirect': '',
        }

        key_words = {
            'text',
            'redirect',
            'http_redirect',
            'template',
            'template_string',
            'file',
            'json',
        }

        # Node response
        if isinstance(raw_response_dict, AbstractNode):
            response_dict['text'] = str(raw_response_dict)

            return response_dict

        # string response
        if isinstance(raw_response_dict, str):
            logger.debug("'%s' is a string based view", view_name)

            response_dict['text'] = raw_response_dict

            return response_dict

        # response dict
        elif isinstance(raw_response_dict, dict):

            # check keys
            if len(set(raw_response_dict) & key_words) != 1:
                raise RuntimeError(
                    'response dicts have to contain exactly one keyword of: {}'.format(  # NOQA: LN001,FS002
                        ', '.join(key_words),
                    ),
                )

            # find keys
            for key in response_dict.keys():
                if key in raw_response_dict:
                    value = raw_response_dict[key]
                    response_dict[key] = value

                    logger.debug(
                        "'%s' sets '%s' to %r", view_name, key, value)

        # redirects
        if 'redirect' in raw_response_dict:
            response_dict['redirect'] = raw_response_dict['redirect']

        # http redirect
        elif 'http_redirect' in raw_response_dict:
            response_dict['http_redirect'] = raw_response_dict['http_redirect']

        # template response
        elif('template' in raw_response_dict or
             'template_string' in raw_response_dict):

            logger.debug("'%s' is a template view", view_name)

            template_context = raw_response_dict

            if 'context' in template_context:
                template_context = template_context['context']

            # template file
            if 'template' in raw_response_dict:
                response_dict['text'] = \
                    self.server.templating_engine.render_template(
                        raw_response_dict['template'],
                        template_context,
                    )

            # template string
            else:
                response_dict['text'] = \
                    self.server.templating_engine.render_string(
                        raw_response_dict['template_string'],
                        template_context,
                    )

        # json response
        elif 'json' in raw_response_dict:
            logger.debug("'%s' is a json view", view_name)

            response_dict['text'] = dumps(raw_response_dict['json'])
            response_dict['content_type'] = 'application/json'

        return response_dict



from __future__ import annotations

from collections.abc import Container, Awaitable
from concurrent.futures import CancelledError
from typing import TYPE_CHECKING
from datetime import datetime
from enum import Enum
import threading
import logging
import asyncio
import time

from typing_extensions import Literal
from yarl import URL

from lona.protocol import (
    encode_input_event_ack,
    encode_http_redirect,
    encode_view_start,
    encode_view_stop,
    encode_redirect,
    encode_data,
    DATA_TYPE,
)
from lona.exceptions import StopReason, ServerStop, UserAbort
from lona.errors import ForbiddenError, ClientError
from lona.html.abstract_node import AbstractNode
from lona.unique_ids import generate_unique_id
from lona.events.input_event import InputEvent
from lona.imports import get_object_repr
from lona.html.document import Document
from lona.request import Request

# avoid import cycles
if TYPE_CHECKING:
    from lona.server import LonaServer


logger = logging.getLogger('lona.view_runtime')
input_events_logger = logging.getLogger('lona.input_events')


class VIEW_RUNTIME_STATE(Enum):
    NOT_STARTED = 10

    RUNNING = 21
    WAITING_FOR_IOLOOP = 22
    SLEEPING = 23
    WAITING_FOR_INPUT = 24

    FINISHED = 31
    CRASHED = 32
    STOPPED_BY_USER = 33
    STOPPED_BY_SERVER = 34


class ViewRuntime:
    def __init__(self, server, url, route, match_info, post_data=None,
                 frontend=False, start_connection=None):

        self.server: LonaServer = server
        self.url = URL(url or '')
        self.route = route
        self.match_info = match_info
        self.post_data = post_data or {}
        self.frontend = frontend
        self.start_connection = start_connection

        # setup request
        self.request = Request(
            view_runtime=self,
            connection=self.start_connection,
        )

        # find view
        if route:
            self.view = route.view

        else:
            self.view = self.server.settings.CORE_ERROR_404_VIEW

        if frontend:
            self.view = self.server.settings.CORE_FRONTEND_VIEW

            if route and route.frontend_view:
                self.view = route.frontend_view

        self.view_class = self.server.view_loader.load(self.view)
        self.name = repr(self.view_class)

        self.view = self.view_class(
            server=self.server,
            view_runtime=self,
            request=self.request,
        )

        # setup state
        self.connections = {}
        # contains: {
        #     connection: (window_id, url),
        # }

        self.document = Document()

        self.stopped: asyncio.Future[Literal[True]] = asyncio.Future(loop=self.server.loop)  # NOQA: LN001
        self.is_stopped = False
        self.stop_reason = None
        self.is_daemon = False

        self.view_runtime_id = generate_unique_id(name_space='view_runtimes')
        self.state = VIEW_RUNTIME_STATE.NOT_STARTED
        self.thread_ident = None
        self.thread_name = None
        self.started_at = None
        self.stopped_at = None

        self.pending_input_events = {
            'event': None,
            'click': None,
            'change': None,
        }

    # state ###################################################################
    @property
    def state(self):
        if self.is_stopped:
            if self.stop_reason is not None:
                if isinstance(self.stop_reason, UserAbort):
                    return VIEW_RUNTIME_STATE.STOPPED_BY_USER

                elif isinstance(self.stop_reason, ServerStop):
                    return VIEW_RUNTIME_STATE.STOPPED_BY_SERVER

                else:
                    return VIEW_RUNTIME_STATE.CRASHED

            else:
                return VIEW_RUNTIME_STATE.FINISHED

        return self._state

    @state.setter
    def state(self, new_state):
        self._state = new_state

    # middlewares #############################################################
    def run_middlewares(self, connection, window_id, url):
        try:
            handled, raw_response_dict, middleware = \
                self.server.middleware_controller.handle_request(
                    view=self.view,
                    request=self.request,
                )

            if handled:
                if not raw_response_dict:
                    raw_response_dict = ''

                self.send_view_start(
                    connections={
                        connection: (window_id, url),
                    },
                )

                response_dict = self.handle_raw_response_dict(
                    raw_response_dict,
                )

                self.send_view_stop(
                    connections={
                        connection: (window_id, url),
                    },
                )

                return response_dict

        except ForbiddenError as exception:
            self.send_view_start(
                connections={
                    connection: (window_id, url),
                },
            )

            view_class = self.server.view_loader.load(
                self.server.settings.CORE_ERROR_403_VIEW,
            )

            view = view_class(
                server=self.server,
                view_runtime=self,
                request=self.request,
            )

            response_dict = self.handle_raw_response_dict(
                view.handle_request(
                    self.request,
                    exception=exception,
                ),
            )

            self.send_view_stop(
                connections={
                    connection: (window_id, url),
                },
            )

            return response_dict

        except Exception as exception:
            logger.exception('Exception raised while running middleware hooks')

            self.send_view_start(
                connections={
                    connection: (window_id, url),
                },
            )

            view_class = self.server.view_loader.load(
                self.server.settings.CORE_ERROR_500_VIEW,
            )

            view = view_class(
                server=self.server,
                view_runtime=self,
                request=self.request,
            )

            return self.handle_raw_response_dict(
                raw_response_dict=view.handle_request(
                    self.request,
                    exception=exception,
                ),
                connections={
                    connection: (window_id, url),
                },
            )

    # start and stop ##########################################################
    def run_shutdown_hook(self):
        # TODO: remove after 1.8

        logger.debug(
            'running %s with stop reason %s',
            self.view.on_shutdown,
            self.stop_reason,
        )

        stop_reason = self.stop_reason

        if not isinstance(stop_reason, (ServerStop, CancelledError)):
            stop_reason = None

        try:
            self.view.on_shutdown(stop_reason)

        except Exception:
            logger.exception(
                'Exception raised while running %s',
                self.view.on_shutdown,
            )

    def run_stop_hook(self):
        logger.debug(
            'running %s with stop reason %s',
            self.view.on_stop,
            self.stop_reason,
        )

        try:
            self.view.on_stop(self.stop_reason)

        except Exception:
            logger.exception(
                'Exception raised while running %s',
                self.view.on_stop,
            )

    def run_cleanup_hook(self):
        from lona.view import LonaView

        # FIXME: this inline import is necessary to avoid circular import
        # this can be fixed by moving runtime state changes from LonaView to
        # ViewRuntime

        def _run():
            logger.debug('running %s', self.view.on_cleanup)

            try:
                self.view.on_cleanup()

            except Exception:
                logger.exception(
                    'Exception raised while running %s',
                    self.view.on_cleanup,
                )

        # if on_cleanup is not defined by the view class
        # it is unnecessary to start a thread
        if self.view_class.on_cleanup is LonaView.on_cleanup:
            return

        self.server.run_function_async(_run)

    def start(self):
        try:
            # update internal state
            current_thread = threading.current_thread()

            self.thread_ident = current_thread.ident
            self.thread_name = current_thread.name
            self.state = VIEW_RUNTIME_STATE.RUNNING
            self.started_at = datetime.now()

            # test CLIENT_VIEW_START_TIMEOUT
            if self.server.settings.TEST_VIEW_START_TIMEOUT:
                time.sleep(self.server.settings.CLIENT_VIEW_START_TIMEOUT + 1)

            # start view
            self.send_view_start()

            # run view
            raw_response_dict = self.view.handle_request(self.request) or ''

            # response dicts
            if isinstance(raw_response_dict, AbstractNode):
                if self.route and not self.route.interactive:
                    raw_response_dict = str(raw_response_dict)

                else:
                    self.view.show(html=raw_response_dict)

                    return

            # check if non-interactive features got used in
            # interactive mode
            if(self.route and self.route.interactive and
                isinstance(raw_response_dict, dict) and (
                    'json' in raw_response_dict or
                    'file' in raw_response_dict)):

                raise RuntimeError(
                    'JSON and file responses are only available in non-interactive mode',
                )

            return self.handle_raw_response_dict(raw_response_dict)

        # view got stopped from outside
        except(StopReason, CancelledError) as exception:
            self.stop_reason = exception

        # 403 Forbidden
        except ForbiddenError as exception:
            self.stop_reason = exception

            view_class = self.server.view_loader.load(
                self.server.settings.CORE_ERROR_403_VIEW,
            )

            view = view_class(
                server=self.server,
                view_runtime=self,
                request=self.request,
            )

            return self.handle_raw_response_dict(
                view.handle_request(
                    self.request,
                    exception=exception,
                ),
            )

        # 500 Internal Error
        except Exception as exception:
            self.stop_reason = exception

            logger.exception('Exception raised while running %s', self.view)

            view_class = self.server.view_loader.load(
                self.server.settings.CORE_ERROR_500_VIEW,
            )

            view = view_class(
                server=self.server,
                view_runtime=self,
                request=self.request,
            )

            return self.handle_raw_response_dict(
                view.handle_request(
                    self.request,
                    exception=exception,
                ),
            )

        finally:
            self.is_stopped = True
            self.stopped_at = datetime.now()
            self.send_view_stop()
            self.run_stop_hook()

        self.run_shutdown_hook()

    def stop(self, reason=UserAbort, clean_up=True):
        self.stop_reason = reason

        if not isinstance(self.stop_reason, Exception):
            self.stop_reason = self.stop_reason()

        # let all calls on request.view.await_sync() crash with
        # the set stop reason
        def _set_stopped():
            if not self.stopped.done():
                self.stopped.set_result(True)

        self.server.loop.call_soon_threadsafe(_set_stopped)

        # cancel all pending user input events
        for pending_input_event in self.pending_input_events.copy().items():
            if not pending_input_event[1]:
                continue

            name, (future, nodes) = pending_input_event

            if not future.done():
                future.set_exception(self.stop_reason)

        # clean up
        if clean_up:
            # drop all connections
            self.connections = {}

            # multi user views have no start_connection
            if self.start_connection:
                self.server.view_runtime_controller.remove_view_runtime(self)

    def issue_500_error(self, exception):
        # stop the runtime but don't run cleanup code to get the
        # output of the 500 handle through
        self.stop(reason=exception, clean_up=False)

        # run 500 view
        self.send_view_start()

        view_class = self.server.view_loader.load(
            self.server.settings.CORE_ERROR_500_VIEW,
        )

        view = view_class(
            server=self.server,
            view_runtime=self,
            request=self.request,
        )

        self.handle_raw_response_dict(
            view.handle_request(
                self.request,
                exception=exception,
            ),
        )

        self.send_view_stop()

    # connection management ###################################################
    def add_connection(self, connection, window_id, url):
        with self.document.lock:
            self.connections[connection] = (window_id, url)

            data_type, data = self.document.serialize()

            if not data:
                return

            self.send_data(
                data=[data_type, data],
                connections={connection: (window_id, url)},
            )

    def reconnect_connection(self, connection, window_id, url):
        with self.document.lock:
            self.connections[connection] = (window_id, url)

            self.send_view_start(
                connections={
                    connection: (window_id, url),
                },
            )

            data_type, data = self.document.serialize()

            if not data:
                return

            self.send_data(
                data=[data_type, data],
                connections={connection: (window_id, url)},
            )

    def remove_connection(self, connection, window_id=None):
        with self.document.lock:
            if connection in self.connections:
                if window_id:
                    if self.connections[connection][0] == window_id:
                        self.connections.pop(connection)

                else:
                    self.connections.pop(connection)

            # if the last connection gets closed and the view should
            # not continue running in background, it gets stopped
            if not self.connections and not self.is_daemon:
                self.stop()

    # lona messages ###########################################################
    def send_redirect(self, target_url, connections=None):
        with self.document.lock:
            connections = connections or self.connections

            for connection, (window_id, _url) in connections.items():
                if not connection.interactive:
                    continue

                message = encode_redirect(
                    window_id=window_id,
                    view_runtime_id=self.view_runtime_id,
                    target_url=target_url,
                )

                connection.send_str(message)

    def send_http_redirect(self, target_url, connections=None):
        with self.document.lock:
            connections = connections or self.connections

            for connection, (window_id, _url) in connections.items():
                if not connection.interactive:
                    continue

                message = encode_http_redirect(
                    window_id=window_id,
                    view_runtime_id=self.view_runtime_id,
                    target_url=target_url,
                )

                connection.send_str(message)

    def _send_html_update(self, title, data, connections=None):
        with self.document.lock:
            connections = connections or self.connections

            data_type, patches = data

            for connection, (window_id, _url) in connections.items():
                if not connection.interactive:
                    continue

                # filter patches
                _patches = []

                for patch in patches:
                    if(patch.issuer and
                       patch.issuer == (connection, window_id)):

                        continue

                    _patches.append(patch.data)

                # send message
                message = encode_data(
                    window_id=window_id,
                    view_runtime_id=self.view_runtime_id,
                    title=title,
                    data=[data_type, _patches],
                )

                connection.send_str(message)

    def send_data(self, title=None, data=None, connections=None):
        if data and data[0] == DATA_TYPE.HTML_UPDATE:
            return self._send_html_update(
                title=title,
                data=data,
                connections=connections or {},
            )

        with self.document.lock:
            connections = connections or self.connections

            # send message
            for connection, (window_id, _url) in connections.items():
                if not connection.interactive:
                    continue

                message = encode_data(
                    window_id=window_id,
                    view_runtime_id=self.view_runtime_id,
                    title=title,
                    data=data,
                )

                connection.send_str(message)

    def send_view_start(self, connections=None):
        with self.document.lock:
            connections = connections or self.connections

            for connection, (window_id, _url) in connections.items():
                if not connection.interactive:
                    continue

                message = encode_view_start(
                    window_id=window_id,
                    view_runtime_id=self.view_runtime_id,
                )

                connection.send_str(message)

    def send_view_stop(self, connections=None):
        with self.document.lock:
            connections = connections or self.connections

            for connection, (window_id, _url) in connections.items():
                if not connection.interactive:
                    continue

                message = encode_view_stop(
                    window_id=window_id,
                    view_runtime_id=self.view_runtime_id,
                )

                connection.send_str(message)

    def handle_raw_response_dict(self, raw_response_dict, connections=None):
        connections = connections or self.connections

        response_dict = self.server.response_parser.render_response_dict(
            raw_response_dict=raw_response_dict,
            view_name=self.name,
        )

        if response_dict['redirect']:
            self.send_redirect(
                response_dict['redirect'],
                connections=connections,
            )

        elif response_dict['http_redirect']:
            self.send_http_redirect(
                response_dict['http_redirect'],
                connections=connections,
            )

        elif response_dict['text']:
            with self.document.lock:
                data = self.document.apply(response_dict['text'])
                self.send_data(data=data, connections=connections)

        return response_dict

    # input events ############################################################
    def await_input_event(
            self,
            nodes: None | Container[AbstractNode] = None,
            event_type: str = 'event',
    ) -> InputEvent:
        async def _await_input_event() -> InputEvent:
            if self.pending_input_events[event_type] is not None:
                raise RuntimeError(f'already waiting for a {event_type} event')

            future: Awaitable[InputEvent] = asyncio.Future()
            self.pending_input_events[event_type] = [future, nodes or []]

            return await future

        return self.server.run_coroutine_sync(_await_input_event())

    def handle_input_event(self, connection, window_id, payload):
        input_events_logger.debug(
            'runtime #%s: handling event #%s',
            self.view_runtime_id,
            payload[0],
        )

        input_event = InputEvent(
            request=self.request,
            payload=payload,
            document=self.document,
            connection=connection,
            window_id=window_id,
        )

        def send_html_patches():
            with self.document.lock:
                data = self.document.apply(self.document.html)

                if data:
                    self.send_data(data=data)

        def log_handled_message():
            input_events_logger.debug(
                'runtime #%s: event #%s: was handled',
                self.view_runtime_id,
                payload[0],
            )

        def run_handler(handler, input_event):
            return_value = handler(input_event)

            if not isinstance(return_value, (InputEvent, dict, type(None))):
                raise RuntimeError(
                    '%r returned an unexpected type (%r)',
                    handler,
                    return_value,
                )

            if isinstance(return_value, dict):
                response_parser = self.server.response_parser

                response_dict = response_parser.parse_event_response_dict(
                    return_value,
                )

                self.handle_raw_response_dict(response_dict)

                return response_dict

            return return_value

        try:
            # sending input event ack
            input_events_logger.debug(
                'runtime #%s: event #%s: sending ACK',
                self.view_runtime_id,
                payload[0],
            )

            # test CLIENT_INPUT_EVENT_TIMEOUT
            if self.server.settings.TEST_INPUT_EVENT_TIMEOUT:
                time.sleep(self.server.settings.CLIENT_INPUT_EVENT_TIMEOUT + 1)

            connection.send_str(
                encode_input_event_ack(
                    window_id=self.connections[connection][0],
                    view_runtime_id=self.view_runtime_id,
                    input_event_id=payload[0],
                ),
            )

            # input event root handler
            input_events_logger.debug(
                'runtime #%s: event #%s: running View.handle_input_event_root()',
                self.view_runtime_id,
                payload[0],
            )

            return_value = run_handler(
                self.view.handle_input_event_root,
                input_event,
            )

            if isinstance(return_value, dict):
                return log_handled_message()

            elif not return_value:
                send_html_patches()

                return log_handled_message()

            else:
                input_event = return_value

            # node input event handler
            for node in input_event.nodes:
                input_events_logger.debug(
                    'runtime #%s: event #%s: running %s.handle_input_event()',
                    self.view_runtime_id,
                    payload[0],
                    get_object_repr(node),
                )

                return_value = run_handler(
                    node.handle_input_event,
                    input_event,
                )

                if isinstance(return_value, dict):
                    return log_handled_message()

                elif not return_value:
                    send_html_patches()

                    return log_handled_message()

                else:
                    input_event = return_value

            # pending input events
            if(input_event.name in self.pending_input_events and
               self.pending_input_events[input_event.name] is not None):

                future, nodes = self.pending_input_events[input_event.name]

                if not nodes or input_event.node in nodes:
                    input_events_logger.debug(
                        'runtime #%s: event #%s: is handled as pending %s event',
                        self.view_runtime_id,
                        payload[0],
                        input_event.name,
                    )

                    future.set_result(input_event)
                    self.pending_input_events[input_event.name] = None

                    return log_handled_message()

            if self.pending_input_events['event'] is not None:
                future, nodes = self.pending_input_events['event']

                if not nodes or input_event.node in nodes:
                    input_events_logger.debug(
                        'runtime #%s: event #%s: is handled as pending generic event',
                        self.view_runtime_id,
                        payload[0],
                    )

                    future.set_result(input_event)
                    self.pending_input_events['event'] = None

                    return log_handled_message()

            # input event handler (class based views)
            input_events_logger.debug(
                'runtime #%s: event #%s: running View.handle_input_event()',
                self.view_runtime_id,
                payload[0],
            )

            return_value = run_handler(
                self.view.handle_input_event,
                input_event,
            )

            if isinstance(return_value, dict):
                return log_handled_message()

            send_html_patches()

            return log_handled_message()

        except(StopReason, CancelledError):
            input_events_logger.debug(
                'runtime #%s: event #%s: handling was stopped',
                self.view_runtime_id,
                payload[0],
            )

        except Exception as exception:
            input_events_logger.exception(
                'runtime #%s: event #%s: exception occurred',
                self.view_runtime_id,
                payload[0],
            )

            self.issue_500_error(exception)

    def handle_client_error(self, connection, window_id, payload):
        logger.error(
            'client error raised:\n%s',
            payload[0],
        )

        self.issue_500_error(ClientError(message=payload[0]))



from __future__ import annotations

from argparse import RawTextHelpFormatter, ArgumentParser, Namespace
from typing import overload, Callable, Any
from tempfile import TemporaryDirectory
from asyncio import AbstractEventLoop
from os import PathLike
import contextlib
import logging
import os

from typing_extensions import Literal
from aiohttp.web import Application

from lona.command_line.run_server import run_server
from lona import default_settings, LonaView
from lona.worker_pool import WorkerPool
from lona.logging import setup_logging
from lona.settings import Settings
from lona.server import LonaServer
from lona.routing import Route

logger = logging.getLogger('lona.app')


class LonaApp:
    def __init__(self, script_path: PathLike | str) -> None:
        self.script_path: PathLike | str = script_path
        self.project_root: str = os.path.dirname(self.script_path)

        self.aiohttp_app: None | Application = None
        self.server: None | LonaServer = None

        # setup tempdir
        self.temp_dir: TemporaryDirectory = TemporaryDirectory()

        self.template_dir: str = os.path.join(
            self.temp_dir.name,
            'templates',
        )

        self.static_dir: str = os.path.join(
            self.temp_dir.name,
            'static',
        )

        os.makedirs(self.template_dir)
        os.makedirs(self.static_dir)

        # setup routes
        self.routes: list[Route] = []

        # setup settings
        self.settings: Settings = Settings()
        self.settings.add(default_settings.__file__)

    # helper ##################################################################
    def _get_settings_as_dict(self) -> dict[str, Any]:
        settings = {}

        for name in self.settings:
            settings[name] = self.settings.get(name)

        return settings

    def resolve_path(self, path: str) -> str:
        if path.startswith('/'):
            return path

        return os.path.normpath(
            os.path.join(
                self.project_root,
                path,
            ),
        )

    # decorator ###############################################################
    def route(
            self,
            # 1 = lona.MATCH_ALL  https://github.com/python/mypy/issues/10026
            raw_pattern: str | Literal[1],
            name: str = '',
            interactive: bool = True,
            http_pass_through: bool = False,
            frontend_view: None | str | LonaView = None,
    ) -> Callable[[type[LonaView]], None]:

        def decorator(view_class: type[LonaView]) -> None:
            self.routes.append(
                Route(
                    raw_pattern=raw_pattern,
                    view=view_class,
                    name=name,
                    interactive=interactive,
                    http_pass_through=http_pass_through,
                    frontend_view=frontend_view,
                ),
            )

        return decorator

    # middleware
    @overload
    def middleware(self) -> Callable[[type], None]:
        ...

    @overload
    def middleware(self, middleware_class: type) -> None:
        ...

    def middleware(self, middleware_class: None | type = None) -> None | Callable[[type], None]:  # NOQA: LN001
        def decorator(middleware_class: type) -> None:
            self.settings.MIDDLEWARES.append(middleware_class)

        if callable(middleware_class):
            decorator(middleware_class)

            return None

        else:
            return decorator

    # frontend
    @overload
    def frontend_view(self) -> Callable[[type[LonaView]], None]:
        ...

    @overload
    def frontend_view(self, view_class: type[LonaView]) -> None:
        ...

    def frontend_view(self, view_class: None | type[LonaView] = None) -> None | Callable[[type[LonaView]], None]:  # NOQA: LN001
        def decorator(view_class: type[LonaView]) -> None:
            self.settings.FRONTEND_VIEW = view_class

        if callable(view_class):
            decorator(view_class)

            return None

        else:
            return decorator

    # files ###################################################################
    def _add_file(
            self,
            temp_dir: str,
            name: str,
            string: str = '',
            path: str = '',
    ) -> None:

        if name.startswith('/'):
            name = name[1:]

        dirname = os.path.join(
            temp_dir,
            os.path.dirname(name),
        )

        with contextlib.suppress(FileExistsError):
            os.makedirs(dirname)

        full_name = os.path.join(
            dirname,
            os.path.basename(name),
        )

        if path:
            _path = self.resolve_path(path)

            if not os.path.exists(_path):
                raise FileNotFoundError(path)

            os.symlink(os.path.abspath(_path), full_name)

        elif string:
            with open(full_name, 'w+') as f:
                f.write(string)

        else:
            raise ValueError('either string or path have to be set')

    def add_template(
            self,
            name: str,
            string: str = '',
            path: str = '',
    ) -> None:

        return self._add_file(
            temp_dir=self.template_dir,
            name=name,
            string=string,
            path=path,
        )

    def add_static_file(
            self,
            name: str,
            string: str = '',
            path: str = '',
    ) -> None:

        return self._add_file(
            temp_dir=self.static_dir,
            name=name,
            string=string,
            path=path,
        )

    # command line ############################################################
    def parse_command_line(self) -> dict:
        from lona.command_line.handle_command_line import (
            parse_overrides,
            DESCRIPTION,
        )

        parser = ArgumentParser(
            prog=str(self.script_path),
            formatter_class=RawTextHelpFormatter,
            description=DESCRIPTION,
        )

        parser.add_argument(
            '-l',
            '--log-level',
            choices=['debug', 'info', 'warn', 'error', 'critical'],
        )

        parser.add_argument(
            '--loggers',
            type=str,
            nargs='+',
        )

        parser.add_argument(
            '--debug-mode',
            choices=['messages', 'views', 'input-events', 'view-events'],
        )

        parser.add_argument(
            '-o',
            '--settings-pre-overrides',
            nargs='+',
        )

        parser.add_argument(
            '-O',
            '--settings-post-overrides',
            nargs='+',
        )

        parser.add_argument(
            '--host',
            type=str,
        )

        parser.add_argument(
            '--port',
            type=int,
        )

        parser.add_argument(
            '--shutdown-timeout',
            type=float,
        )

        parser.add_argument(
            '--shell',
            action='store_true',
        )

        parser.add_argument(
            '--shell-server-url',
            type=str,
        )

        args = vars(parser.parse_args())

        for key, value in args.copy().items():
            if not value:
                args.pop(key)

        if 'settings_pre_overrides' in args:
            args['settings_pre_overrides'] = parse_overrides(
                args['settings_pre_overrides'],
            )

        if 'settings_post_overrides' in args:
            args['settings_post_overrides'] = parse_overrides(
                args['settings_post_overrides'],
            )

        return args

    # server ##################################################################
    def setup_server(
            self,
            loop: None | AbstractEventLoop = None,
            settings_pre_overrides: None | dict[str, Any] = None,
            settings_post_overrides: None | dict[str, Any] = None,
    ) -> None:

        # finish settings
        self.settings.CORE_TEMPLATE_DIRS.insert(0, self.template_dir)
        self.settings.STATIC_DIRS.insert(0, self.static_dir)

        # setup server
        self.aiohttp_app = Application(loop=loop)

        settings_post_overrides = {
            **self._get_settings_as_dict(),
            **(settings_post_overrides or {}),
        }

        self.server = LonaServer(
            app=self.aiohttp_app,
            project_root=self.project_root,
            settings_pre_overrides=settings_pre_overrides,
            settings_post_overrides=settings_post_overrides,
            routes=self.routes,
        )

        # setup worker pool
        worker_pool = WorkerPool(
            settings=self.server.settings,
        )

        self.server.set_loop(loop)
        self.server.set_worker_pool(worker_pool)

    def run(
            self,
            loop: None | AbstractEventLoop = None,
            parse_command_line: bool = True,
            **args: Any,
    ) -> None:

        # setup arguments
        server_args = Namespace(
            host='localhost',
            port=8080,
            shell_server_url='',
            shutdown_timeout=0,
            log_level='info',
            loggers=[],
            debug_mode='',
            shell=False,
            settings_pre_overrides=None,
            settings_post_overrides=None,
        )

        for key, value in args.items():
            setattr(server_args, key, value)

        if parse_command_line:
            command_line_args = self.parse_command_line()

            for key, value in command_line_args.items():
                setattr(server_args, key, value)

        setup_logging(server_args)

        # setup server
        self.setup_server(
            loop=loop,
            settings_pre_overrides=server_args.settings_pre_overrides,
            settings_post_overrides=server_args.settings_post_overrides,
        )

        # start server
        run_server(
            args=server_args,
            app=self.aiohttp_app,
            server=self.server,
        )



from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING
from copy import copy
import logging
import os

from lona.static_files import StyleSheet, StaticFile, Script
from lona.imports import get_all_subclasses, get_file
from lona.html.abstract_node import AbstractNode

# avoid import cycles
if TYPE_CHECKING:
    from lona.server import LonaServer


logger = logging.getLogger('lona.static_file_loader')


class StaticFileLoader:
    def __init__(self, server: LonaServer) -> None:
        self.server: LonaServer = server

        self.static_dirs: list[str] = [
            *self.server.settings.STATIC_DIRS,
            *self.server.settings.CORE_STATIC_DIRS,
        ]

        # resolving potential relative paths
        for index, static_dir in enumerate(self.static_dirs):
            self.static_dirs[index] = self.server.resolve_path(static_dir)

        logger.debug('static dirs %s loaded', repr(self.static_dirs)[1:-1])

        # discover
        self.stylesheets: list[StyleSheet] = []
        self.scripts: list[Script] = []
        self.others: list[StaticFile] = []
        self.static_files: list[StaticFile] = self._discover_static_files()

        logger.debug('%d stylesheets discovered', len(self.stylesheets))
        logger.debug('%d scripts discovered', len(self.scripts))

        # render html files
        logger.debug('rendering html files')

        # stylesheets
        self.style_tags_html: str = self.server.templating_engine.render_template(  # NOQA: LN001
            self.server.settings.STATIC_FILES_STYLE_TAGS_TEMPLATE,
            {
                'stylesheets': [i for i in self.stylesheets
                                if i.enabled and i.link],
            },
        )

        # scripts
        self.script_tags_html: str = self.server.templating_engine.render_template(  # NOQA: LN001
            self.server.settings.STATIC_FILES_SCRIPT_TAGS_TEMPLATE,
            {
                'scripts': [i for i in self.scripts
                            if i.enabled and i.link],
            },
        )

    def _iter_all_declarations(self) -> Iterator[tuple[type, Iterator[StaticFile]]]:  # NOQA: LN001
        logger.debug('discover node classes')
        node_classes = [AbstractNode] + list(get_all_subclasses(AbstractNode))
        logger.debug('%d node classes discovered', len(node_classes))

        for node_class in node_classes:
            yield node_class, iter(node_class.STATIC_FILES)

        logger.debug('discover view classes')
        view_classes = self.server.view_loader.get_all_views()
        logger.debug('%d view classes discovered', len(view_classes))

        for view_class in view_classes:
            if not hasattr(view_class, 'STATIC_FILES'):
                continue

            yield view_class, iter(view_class.STATIC_FILES)

    def _discover_static_files(self) -> list[StaticFile]:
        discovered_names: set[str] = set()
        # can't use set because StaticFileDecl is unhashable
        visited_declarations: list[StaticFile] = []

        for cls, declarations in self._iter_all_declarations():
            class_path = get_file(cls)
            class_dirname = os.path.dirname(class_path)
            context = f'{class_path}::{cls.__name__}'
            for file_declaration in declarations:
                static_file = self._process_declaration(
                    file_declaration,
                    context,
                    class_dirname,
                    visited_declarations,
                    discovered_names,
                )

                if static_file is None:
                    continue

                # sort static file into the right cache
                if isinstance(static_file, StyleSheet):
                    self.stylesheets.append(static_file)

                elif isinstance(static_file, Script):
                    self.scripts.append(static_file)

                else:
                    self.others.append(static_file)

        # sort
        self.stylesheets.sort(key=lambda x: x.sort_order.value)
        self.scripts.sort(key=lambda x: x.sort_order.value)
        self.others.sort(key=lambda x: x.sort_order.value)

        return [
            *self.stylesheets,
            *self.scripts,
            *self.others,
        ]

    def _process_declaration(
            self,
            file_declaration: StaticFile,
            context: str,
            dirname: str,
            visited: list[StaticFile],
            discovered_names: set[str],
    ) -> None | StaticFile:
        if not isinstance(file_declaration, StaticFile):
            logger.error(
                '%s: unknown type found: %r',
                context,
                file_declaration,
            )

            return None

        if file_declaration in visited:
            return None

        visited.append(file_declaration)

        if file_declaration.name in discovered_names:
            logger.error(
                "%s: static file with name '%s' already used",
                context,
                file_declaration.name,
            )

            return None

        # create a local copy
        static_file = copy(file_declaration)

        static_file.static_url_prefix = self.server.settings.STATIC_URL_PREFIX

        # patch path
        static_file_rel_path = static_file.path

        static_file_abs_path = os.path.join(dirname, static_file_rel_path)

        if not os.path.exists(static_file_abs_path):
            logger.error(
                "%s: path '%s' does not exist",
                context,
                static_file_abs_path,
            )

            return None

        static_file.path = static_file_abs_path

        # patch url
        if not static_file.url:
            url = static_file_rel_path

        else:
            url = static_file.url

        if url.startswith('/'):
            url = url[1:]

        static_file.url = url

        # disable static files that are not enabled by default and
        # are not configured to be enabled
        if(not static_file.enabled_by_default and
           static_file.name not in self.server.settings.STATIC_FILES_ENABLED):
            static_file.enabled = False

        # disable static files that are disabled explicitly
        if static_file.name in self.server.settings.STATIC_FILES_DISABLED:
            static_file.enabled = False

        discovered_names.add(static_file.name)

        return static_file

    def resolve_path(self, path: str) -> None | str:
        logger.debug("resolving '%s'", path)

        rel_path = path

        if rel_path.startswith('/'):
            rel_path = rel_path[1:]

        # javascript client
        client_url = self.server.settings.STATIC_FILES_CLIENT_URL

        if client_url.startswith('/'):
            client_url = client_url[1:]

        if rel_path == client_url:
            logger.debug('returning javascript client')

            return self.server.client_pre_compiler.resolve()

        # searching in static dirs
        for static_dir in self.static_dirs:
            abs_path = os.path.join(static_dir, rel_path)

            logger.debug("trying '%s'", abs_path)

            if os.path.exists(abs_path):
                if os.path.isdir(abs_path):
                    logger.debug(
                        "'%s' is directory. resolving stopped", abs_path)

                    return None

                logger.debug("returning '%s'", abs_path)

                return abs_path

        # searching in node static files
        for static_file in self.static_files:
            if not static_file.enabled:
                continue

            if static_file.url == path:
                return static_file.path

        logger.debug("'%s' was not found", path)

        return None



from threading import Lock


class UniqueIDGenerator:
    def __init__(self):
        self._lock = Lock()
        self._value = 0

    def __call__(self):
        with self._lock:
            self._value += 1

            return str(self._value)


_name_spaces = {
    '': UniqueIDGenerator(),
    'nodes': UniqueIDGenerator(),
    'view_runtimes': UniqueIDGenerator(),
}


def generate_unique_id(name_space=''):
    return _name_spaces[name_space]()



from pprint import pformat

from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table

from lona import VERSION_STRING


class LonaInfoCommand:
    """
    Show Lona info
    """

    NAME = 'lona_info'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='lona_info')

        argument_parser.parse_args(argv[1:])

        # write version string
        rows = [
            ['Key', 'Value'],
            ['Lona version', f'v{VERSION_STRING}'],
            ['project root', self.repl.locals['server'].project_root],
            ['cli args', pformat(self.repl.locals.get('cli_args', {}))],
        ]

        write_table(rows, self.repl.write)



from rlpython.utils.argument_parser import ReplArgumentParser


class LonaMiddlewaresCommand:
    """
    List all middleware hooks
    """

    NAME = 'lona_middlewares'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(
            repl=self.repl,
            prog='lona_middlewares',
            description=self.__doc__,
        )

        argument_parser.parse_args(argv[1:])

        server = self.repl.locals['server']
        middleware_controller = server.middleware_controller

        # list middlewares
        hook_names = [
            'on_startup',
            'on_shutdown',
            'handle_connection',
            'handle_websocket_message',
            'handle_request',
        ]

        for hook_name in hook_names:
            self.repl.write(f'{hook_name}\n')

            for middleware, _hook in middleware_controller.hooks[hook_name]:
                self.repl.write(
                    f'    {middleware.__module__}.{middleware.__class__.__name__}\n',
                )

            self.repl.write('\n')



import os

from rlpython.utils.argument_parser import ReplArgumentParser
from jinja2.exceptions import TemplateNotFound


class LonaTemplatesCommand:
    """
    List all templates
    """

    NAME = 'lona_templates'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(
            repl=self.repl,
            prog='lona_templates',
            description=self.__doc__,
        )

        argument_parser.add_argument('template-dir', nargs='?')

        argument_parser.add_argument(
            '-l',
            '--list-directories',
            action='store_true',
        )

        argument_parser.add_argument(
            '-r',
            '--resolve',
        )

        args = argument_parser.parse_args(argv[1:])

        server = self.repl.locals['server']

        # resolve
        if args.resolve:
            try:
                template = server.templating_engine.get_template(args.resolve)

            except TemplateNotFound:
                return 1

            self.repl.write(f'{template.filename}\n')

            return

        # list directories
        if args.list_directories:
            for static_dir in server.templating_engine.template_dirs:
                self.repl.write(f'{static_dir}\n')

            return

        # list static files
        templates = {}

        for template_dir in server.templating_engine.template_dirs:
            for root, _dirs, files in os.walk(template_dir):
                for _file in files:
                    name = os.path.join(root, _file)

                    if name in templates:
                        continue

                    templates[name] = template_dir

        template_list = []

        for name, static_dir in templates.items():
            template_list.append(os.path.join(static_dir, name))

        template_list.sort()

        for static_file in template_list:
            self.repl.write(f'{static_file}\n')



from lona.events.event_types import CHANGE
from lona.html.node import Node


class Option(Node):
    TAG_NAME = 'option'


class Select(Node):
    TAG_NAME = 'select'
    EVENTS = [CHANGE]

    def __init__(self, *args, values=None, disabled=False, multiple=False,
                 readonly=False, bubble_up=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.values = values or []
        self.disabled = disabled
        self.multiple = multiple
        self.readonly = readonly
        self.bubble_up = bubble_up

    def handle_input_event(self, input_event):
        if input_event.name == 'change':
            self.value = input_event.data

            input_event = self.handle_change(input_event)

        elif input_event.name == 'click':
            input_event = self.handle_click(input_event)

        else:
            return input_event

        if self.bubble_up:
            return input_event

    # properties ##############################################################
    # disabled
    @property
    def disabled(self):
        return 'disabled' in self.attributes

    @disabled.setter
    def disabled(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('disabled is a boolean property')

        if new_value:
            self.attributes['disabled'] = ''

        else:
            del self.attributes['disabled']

    # multiple
    @property
    def multiple(self):
        return 'multiple' in self.attributes

    @multiple.setter
    def multiple(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('multiple is a boolean property')

        if new_value:
            self.attributes['multiple'] = ''

        else:
            del self.attributes['multiple']

    # readonly
    @property
    def readonly(self):
        return 'readonly' in self.attributes

    @readonly.setter
    def readonly(self, new_value):
        if not isinstance(new_value, bool):
            raise TypeError('readonly is a boolean property')

        if new_value:
            self.attributes['readonly'] = ''

        else:
            del self.attributes['readonly']

    # values
    @property
    def values(self):
        with self.lock:
            values = []

            for node in self.nodes:
                if node.tag_name != 'option':
                    continue

                values.append(
                    (node.attributes.get('value', ''),
                     node.get_text(),
                     'selected' in node.attributes),
                )

            return values

    @values.setter
    def values(self, new_values):
        with self.lock:
            self.clear()

            for i in new_values:
                value, name, selected = (list(i) + [False])[0:3]
                option_node = Option(str(name), value=str(value))

                if selected:
                    option_node.attributes['selected'] = ''

                self.append(option_node)

    # value
    @property
    def value(self):
        with self.lock:
            value = []

            for option in self.nodes:
                if 'selected' in option.attributes:
                    value.append(option.attributes['value'])

            if not value and self.nodes:
                option = self.nodes[0]

                value.append(option.attributes['value'])

            if not value:
                return None

            if 'multiple' not in self.attributes:
                value = value.pop()

            return value

    @value.setter
    def value(self, new_value):
        if not isinstance(new_value, list):
            new_value = [new_value]

        with self.lock:
            for option in self.nodes:
                if option.attributes['value'] in new_value:
                    option.attributes['selected'] = ''

                else:
                    del option.attributes['selected']



class Patch:
    def __init__(self, node_id, patch_type, operation, payload, issuer=None):
        self.issuer = issuer

        self.data = [
            node_id,
            patch_type,
            operation,
            *payload,
        ]

    def __repr__(self):
        return f'<Patch({self.data[0]}, {self.data[1]}, {self.data[2]})>'


class PatchStack:
    def __init__(self):
        self.patches = []

    def add_patch(self, *args, **kwargs):
        self.patches.append(Patch(*args, **kwargs))

    def has_patches(self):
        return bool(self.patches)

    def get_patches(self):
        patch_data = []

        for patch in self.patches:
            patch_data.append(patch)

        return patch_data

    def clear(self):
        self.patches.clear()



from enum import Enum
import json

SEPARATORS = (',', ':')


def default(value):
    # this function does not use isinstance() to avoid import loops

    if isinstance(value, Enum):
        return value.value

    if value.__class__.__name__ == 'WidgetData':
        return value._data

    raise TypeError


def dumps(*args, default=default, separators=SEPARATORS, **kwargs):
    return json.dumps(*args, default=default, separators=separators, **kwargs)



import logging

from lona.protocol import (
    decode_message,
    encode_pong,
    EXIT_CODE,
    PROTOCOL,
    METHOD,
)

logger = logging.getLogger('lona.protocol')


class LonaMessageMiddleware:
    def handle_websocket_message(self, data):
        if not data.message.startswith(PROTOCOL.MESSAGE_PREFIX.value):
            return data

        exit_code, window_id, view_runtime_id, method, payload = \
            decode_message(data.message)

        if exit_code != EXIT_CODE.SUCCESS:
            logger.error('invalid lona message received: %s', data.message)

            return data

        if method == METHOD.PING:
            data.connection.send_str(encode_pong(), wait=False)

            return

        data.server.view_runtime_controller.handle_lona_message(
            connection=data.connection,
            window_id=window_id,
            view_runtime_id=view_runtime_id,
            method=method,
            payload=payload,
        )



from traceback import format_exception
from textwrap import indent
import threading
import datetime
import logging
import socket

from lona.command_line.terminal import (
    terminal_supports_colors,
    colors_are_enabled,
)


class LogFilter(logging.Filter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.excluded = []
        self.included = []

    def clear(self):
        self.excluded.clear()
        self.included.clear()

    def include(self, logger_name):
        self.included.append(logger_name)

    def exclude(self, logger_name):
        self.excluded.append(logger_name)

    def filter(self, record):
        # filter exceptions that lona.command_line.run_server.run_server
        # handles it self
        if record.exc_info:

            # OSErrors
            if(isinstance(record.exc_info[1], OSError) and
               record.exc_info[1].errno in (13, 98)):

                return False

            # socket.gaierror
            if(isinstance(record.exc_info[1], socket.gaierror) and
               record.exc_info[1].errno in (-2, )):

                return False

        if record.name == 'lona':
            # The lona root logger is used by the command line tools
            # to report errors, for example when startup is not possible due
            # an invalid host or port.
            # These errors can't be ignored.

            return True

        if record.name in self.excluded:
            return False

        if self.included and record.name not in self.included:
            return False

        return True


class LogFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.colors_enabled = (
            terminal_supports_colors() and
            colors_are_enabled()
        )

    def format(self, record):
        current_thread_name = threading.current_thread().name

        # format record string
        time_stamp = datetime.datetime.fromtimestamp(record.created)
        time_stamp_str = time_stamp.strftime('%H:%M:%S.%f')

        record_string = '{}{} {}{} {} {} {}'.format(  # NOQA: FS002
            current_thread_name,
            (30 - len(current_thread_name)) * ' ',

            record.levelname,
            (8 - len(record.levelname)) * ' ',

            time_stamp_str,
            record.name,
            record.getMessage(),
        )

        # format exc_info
        if record.exc_info:
            record_string = '{}\n{}'.format(  # NOQA: FS002
                record_string,
                indent(
                    ''.join(format_exception(*record.exc_info))[:-1],
                    prefix='  ',
                ),
            )

        # colors
        if record.levelname == 'INFO' or not self.colors_enabled:
            return record_string

        RED = '31'
        YELLOW = '33'
        WHITE = '37'
        GREEN = '32'
        BACKGROUND_RED = '41'

        BRIGHT = '1'

        style = ''
        background = ''
        color = ''

        if record.levelname == 'DEBUG':
            style = ''
            background = ''
            color = GREEN

        elif record.levelname == 'WARNING':
            style = ''
            background = ''
            color = YELLOW

        elif record.levelname == 'ERROR':
            style = BRIGHT
            background = ''
            color = RED

        elif record.levelname == 'CRITICAL':
            style = BRIGHT
            background = BACKGROUND_RED
            color = WHITE

        if style:
            color = f';{color}'

        if background and color:
            background = f';{background}'

        return f'\033[{style}{color}{background}m{record_string}\033[00m'


def setup_logging(args, log_formatter=None, log_filter=None):
    # set log level
    if not args.debug_mode:
        log_level = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warn': logging.WARN,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
        }[args.log_level.lower()]

    else:
        log_level = logging.DEBUG

    logging.basicConfig(level=log_level)

    # setup log formatting and log filtering
    log_formatter = log_formatter or LogFormatter()
    log_filter = log_filter or LogFilter()

    for handler in logging.getLogger().root.handlers:
        handler.setFormatter(log_formatter)
        handler.addFilter(log_filter)

    if args.debug_mode:
        if args.debug_mode == 'messages':
            log_filter.include('lona.server.websockets')

        elif args.debug_mode == 'views':
            log_filter.include('lona.views')

        elif args.debug_mode == 'input-events':
            log_filter.include('lona.input_events')

        elif args.debug_mode == 'view-events':
            log_filter.include('lona.view_events')

    elif args.loggers:
        for logger_name in args.loggers:
            if logger_name.startswith('_'):
                log_filter.exclude(logger_name[1:])

            else:
                if logger_name.startswith('+'):
                    logger_name = logger_name[1:]

                log_filter.include(logger_name)

    return log_formatter, log_filter



from functools import lru_cache
import logging
import re

from lona import default_settings

ABSTRACT_ROUTE_RE = re.compile(r'<(?P<name>[^:>]+)(:(?P<pattern>[^>]+))?>')
ROUTE_PART_FORMAT_STRING = r'(?P<{}>{})'
DEFAULT_PATTERN = r'[^/]+'
OPTIONAL_TRAILING_SLASH_PATTERN = r'(/)'

MATCH_ALL = 1

logger = logging.getLogger('lona.routing')


class Route:
    def __init__(self, raw_pattern, view, name='', interactive=True,
                 http_pass_through=False, frontend_view=None):

        self.raw_pattern = raw_pattern
        self.view = view
        self.name = name
        self.interactive = interactive
        self.http_pass_through = http_pass_through
        self.frontend_view = frontend_view

        self.path = None
        self.format_string = ''
        self.optional_trailing_slash = False

        # match all
        if self.raw_pattern == MATCH_ALL:
            self.path = MATCH_ALL

        # string or regex
        else:
            if self.raw_pattern.endswith(OPTIONAL_TRAILING_SLASH_PATTERN):
                self.optional_trailing_slash = True

            groups = ABSTRACT_ROUTE_RE.findall(self.raw_pattern)

            # path is no pattern but simple string
            if not groups:
                self.path = self.raw_pattern
                self.format_string = self.raw_pattern

                if self.optional_trailing_slash:
                    suffix_len = len(OPTIONAL_TRAILING_SLASH_PATTERN) * -1

                    self.path = self.path[:suffix_len]
                    self.format_string = self.path[:suffix_len]

                return

            pattern_names = [i[0] for i in groups]
            patterns = [(i[0], i[2] or DEFAULT_PATTERN) for i in groups]
            cleaned_pattern = ABSTRACT_ROUTE_RE.sub('{}', self.raw_pattern)

            # setup format string
            self.format_string = cleaned_pattern.format(
                *['{' + i + '}' for i in pattern_names])

            # compile pattern
            self.pattern = re.compile(
                r'^{}{}$'.format(  # NOQA: FS002
                    cleaned_pattern.format(
                        *[ROUTE_PART_FORMAT_STRING.format(*i)
                          for i in patterns],
                    ),
                    (OPTIONAL_TRAILING_SLASH_PATTERN
                        if self.optional_trailing_slash else ''),
                ),
            )

    def match(self, path):
        # match all
        if self.path == MATCH_ALL:
            return True, {}

        # simple string
        if self.path:
            if self.optional_trailing_slash and path.endswith('/'):
                path = path[:-1]

            return path == self.path, {}

        # pattern
        match_object = self.pattern.match(path)

        if not match_object:
            return False, {}

        return True, match_object.groupdict()

    def __repr__(self):
        raw_pattern = self.raw_pattern

        if raw_pattern == MATCH_ALL:
            raw_pattern = 'MATCH_ALL'

        return f'<Route({raw_pattern}, {self.view})>'


class Router:
    def __init__(self):
        self.routes = []

        self.resize_resolve_cache(
            default_settings.ROUTING_RESOLVE_CACHE_MAX_SIZE,
        )

        self.resize_reverse_cache(
            default_settings.ROUTING_REVERSE_CACHE_MAX_SIZE,
        )

    # caches ##################################################################
    def resize_resolve_cache(self, max_size):
        self._resolve_lru_cache = lru_cache(max_size)(self._resolve)

    def resize_reverse_cache(self, max_size):
        self._reverse_lru_cache = lru_cache(max_size)(self._reverse)

    def get_resolve_cache_info(self):
        return self._resolve_lru_cache.cache_info()

    def get_reverse_cache_info(self):
        return self._reverse_lru_cache.cache_info()

    def clear_resolve_cache_info(self):
        return self._resolve_lru_cache.cache_clear()

    def clear_reverse_cache_info(self):
        return self._reverse_lru_cache.cache_clear()

    # routes ##################################################################
    def add_route(self, route):
        self.routes.append(route)

    def add_routes(self, *routes):
        for route in routes:
            self.add_route(route)

    # resolve #################################################################
    def _resolve(self, path):
        logger.debug("resolving '%s'", path)

        for route in self.routes:
            match, match_info = route.match(path)

            if match:
                logger.debug('%s matched', route)

                return True, route, match_info

        logger.debug("no match for '%s'", path)

        return False, None, {}

    def resolve(self, *args, **kwargs):
        return self._resolve_lru_cache(*args, **kwargs)

    # reverse #################################################################
    def _reverse(self, name, *args, **kwargs):
        route = None

        for i in self.routes:
            if i.name == name:
                route = i

                break

        if not route:
            raise ValueError(f"no route named '{name}' found")

        if route.path:
            return route.path

        try:
            return route.format_string.format(*args, **kwargs)

        except KeyError as e:
            key_error = e

        # raise is outside of except block to avoid stacking tracebacks
        raise ValueError(f'missing URL arg: {key_error.args[0]}')

    def reverse(self, *args, **kwargs):
        return self._reverse_lru_cache(*args, **kwargs)



from dataclasses import dataclass
from enum import Enum


class SORT_ORDER(Enum):
    FRAMEWORK = 100
    LIBRARY = 200
    APPLICATION = 300


@dataclass
class StaticFile:
    name: str
    path: str
    url: str = ''
    sort_order: SORT_ORDER = SORT_ORDER.APPLICATION
    link: bool = True
    enabled_by_default: bool = True

    # this is only for internal use
    enabled: bool = True
    # this is only for the object representation and is meant for debugging
    static_url_prefix: str = ''


class StyleSheet(StaticFile):
    pass


class Script(StaticFile):
    pass



from __future__ import annotations

from typing import Any
import os

MAX_WORKER_THREADS = 4
MAX_STATIC_THREADS = 4
MAX_RUNTIME_THREADS = 6

# routing
ROUTING_TABLE = 'lona.default_routes.routes'
ROUTING_RESOLVE_CACHE_MAX_SIZE = 1000
ROUTING_REVERSE_CACHE_MAX_SIZE = 1000

# templating
CORE_TEMPLATE_DIRS = [
    os.path.join(os.path.dirname(__file__), 'templates'),
]

TEMPLATE_DIRS: list[str] = []

FRONTEND_TEMPLATE = 'lona/frontend.html'
ERROR_403_TEMPLATE = 'lona/403.html'
ERROR_404_TEMPLATE = 'lona/404.html'
ERROR_500_TEMPLATE = 'lona/500.html'

TEMPLATE_EXTRA_CONTEXT: dict[str, Any] = {}

# static files
CORE_STATIC_DIRS = [
    os.path.join(os.path.dirname(__file__), 'static'),
]

STATIC_DIRS: list[str] = []
STATIC_URL_PREFIX = '/static/'
STATIC_FILES_SERVE = True
STATIC_FILES_STYLE_TAGS_TEMPLATE = 'lona/style_tags.html'
STATIC_FILES_SCRIPT_TAGS_TEMPLATE = 'lona/script_tags.html'
STATIC_FILES_ENABLED: list[str] = []
STATIC_FILES_DISABLED: list[str] = []
STATIC_FILES_CLIENT_URL = '/lona/lona.js'

# client
CLIENT_RECOMPILE = False
CLIENT_VIEW_START_TIMEOUT = 2
CLIENT_INPUT_EVENT_TIMEOUT = 2
CLIENT_PING_INTERVAL = 60


# sessions
SESSIONS = True
SESSIONS_KEY_GENERATOR = 'lona.middlewares.sessions.generate_session_key'
SESSIONS_KEY_NAME = 'sessionid'
SESSIONS_KEY_RANDOM_LENGTH = 28

# views
CORE_FRONTEND_VIEW = 'lona.default_views.FrontendView'
FRONTEND_VIEW = ''

# error views
CORE_ERROR_403_VIEW = 'lona.default_views.Error403View'
CORE_ERROR_404_VIEW = 'lona.default_views.Error404View'
CORE_ERROR_500_VIEW = 'lona.default_views.Error500View'

ERROR_403_VIEW = ''
ERROR_404_VIEW = ''
ERROR_500_VIEW = ''

# middlewares
CORE_MIDDLEWARES = [
    'lona.middlewares.sessions.LonaSessionMiddleware',
    'lona.middlewares.lona_messages.LonaMessageMiddleware',
]

MIDDLEWARES: list[str] = []

# shell
CORE_COMMANDS = [
    'lona.shell.commands.lona_server_state.LonaServerStateCommand',
    'lona.shell.commands.lona_static_files.LonaStaticFilesCommand',
    'lona.shell.commands.lona_connections.LonaConnectionsCommand',
    'lona.shell.commands.lona_middlewares.LonaMiddlewaresCommand',
    'lona.shell.commands.lona_templates.LonaTemplatesCommand',
    'lona.shell.commands.lona_settings.LonaSettingsCommand',
    'lona.shell.commands.lona_routes.LonaRoutesCommand',
    'lona.shell.commands.lona_views.LonaViewsCommand',
    'lona.shell.commands.lona_info.LonaInfoCommand',
]

COMMANDS: list[str] = []

# testing
TEST_VIEW_START_TIMEOUT = False
TEST_INPUT_EVENT_TIMEOUT = False



from rlpython import embed


def load_commands(server):
    import_strings = server.settings.CORE_COMMANDS + server.settings.COMMANDS
    commands = []

    for import_string in import_strings:
        commands.append(server.acquire(import_string))

    return commands


def embed_shell(server, **embed_kwargs):
    embed_kwargs['commands'] = load_commands(server)

    embed(**embed_kwargs)


def generate_shell_server(server, **embed_kwargs):
    embed_kwargs['commands'] = load_commands(server)
    embed_kwargs['multi_session'] = True

    return embed(**embed_kwargs)



from __future__ import annotations

from collections.abc import AsyncIterator, Iterator, Callable
from typing import AsyncContextManager, cast
from asyncio import AbstractEventLoop, sleep
from contextlib import asynccontextmanager
from dataclasses import dataclass
import time

from aiohttp.test_utils import TestClient
from aiohttp.web import Application
import pytest

from lona.worker_pool import WorkerPool
from lona.server import LonaServer
from lona import LonaApp


@dataclass
class LonaContext:
    client: TestClient
    app: None | LonaApp
    server: LonaServer

    def make_url(self, path: str = '') -> str:
        return str(self.client.make_url(path))


@pytest.fixture()
def lona_app_context(request, aiohttp_client):
    async def setup_lona_app_context(
            setup_app: Callable[[LonaApp], None],
            project_root: str = '',
    ) -> LonaContext:
        # setup lona app
        lona_app = LonaApp(project_root or str(request.fspath))
        setup_app(lona_app)

        # setup client
        def setup_aiohttp_app(loop: AbstractEventLoop) -> Application:
            lona_app.setup_server(loop=loop)

            return cast(Application, lona_app.aiohttp_app)

        client = await aiohttp_client(setup_aiohttp_app)

        return LonaContext(
            client=client,
            app=lona_app,
            server=cast(LonaServer, lona_app.server),
        )

    return setup_lona_app_context


@pytest.fixture()
def lona_project_context(request, aiohttp_client):
    async def setup_lona_project_context(
        project_root: str = '',
        settings: None | list[str] = None,
        settings_pre_overrides: None | dict = None,
        settings_post_overrides: None | dict = None,
    ) -> LonaContext:

        # setup aiohttp app
        server = None

        def setup_aiohttp_app(loop: AbstractEventLoop) -> Application:
            nonlocal server

            aiohttp_app = Application(loop=loop)

            # setup lona server
            server = LonaServer(
                app=aiohttp_app,
                project_root=project_root or request.fspath,
                settings_paths=settings or [],
                settings_pre_overrides=settings_pre_overrides or {},
                settings_post_overrides=settings_post_overrides or {},
            )

            server.set_loop(loop)
            server.set_worker_pool(WorkerPool(settings=server.settings))

            return aiohttp_app

        client = await aiohttp_client(setup_aiohttp_app)

        return LonaContext(
            client=client,
            app=None,
            server=cast(LonaServer, server),
        )

    return setup_lona_project_context


def eventually(timeout: float = 5, interval: float = 1) -> Iterator[AsyncContextManager]:  # NOQA: LN001
    """
    Wait for expected state in async test.

    The function is meant to be iterated. Each time it returns new attempt.
    Attempt is an async context manager to wrap assertions.
    It suppresses all exceptions until time is out.
    If no exception is raised iteration stops.

    Example::

        counter = 0
        for attempt in eventually():
            async with attempt:
                counter += 1
                assert counter > 3

    :param timeout: time in seconds during which it produces new attempts
    :param interval: seconds to sleep between attempts
    """

    end_time = time.time() + timeout
    success = False

    while not success:
        context_manager_was_used = False

        @asynccontextmanager
        async def attempt() -> AsyncIterator[None]:
            nonlocal context_manager_was_used, success
            context_manager_was_used = True

            try:
                yield  # execute assertions

            except Exception:
                if time.time() > end_time:
                    raise
                else:
                    await sleep(interval)

            else:
                success = True

        yield attempt()

        if not context_manager_was_used:
            raise SyntaxError('context manager must be used')



import logging
import asyncio

logger = logging.getLogger('lona.middlewares')


class MiddlewareData:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        values = ', '.join(f'{k}={v!r}' for k, v in self.kwargs.items())
        return f'<MiddlewareData({values})>'


class MiddlewareController:
    HOOKS = [
        # contains: (HOOK_NAME, HOOK_IS_ASYNC, )
        ('on_startup',                      True),
        ('on_shutdown',                     True),
        ('handle_connection',              False),
        ('handle_websocket_message',       False),
        ('handle_request',                 False),
    ]

    def __init__(self, server):
        self.server = server

        self.setup()

    def setup(self):
        # setup middlewares
        logger.debug('setup middlewares')

        self.middleware_classes = (
            self.server.settings.CORE_MIDDLEWARES +
            self.server.settings.MIDDLEWARES
        )

        self.middlewares = []

        for i in self.middleware_classes:
            logger.debug('loading %s', i)

            try:
                if isinstance(i, str):
                    middleware_class = self.server.acquire(i)

                else:
                    middleware_class = i

            except Exception:
                logger.exception('Exception raised while loading %s', i)

                continue

            try:
                self.middlewares.append(middleware_class())

            except Exception:
                logger.exception('Exception raised while initializing %s', i)

        # discover middleware hooks
        logger.debug('discover middleware hooks')

        self.hooks = {
            hook_name: [] for hook_name, _ in self.HOOKS
        }

        for middleware in self.middlewares:
            logger.debug('discovering %s', middleware)

            for hook_name, hook_is_async in self.HOOKS:
                if not hasattr(middleware, hook_name):
                    continue

                hook = getattr(middleware, hook_name)

                if not callable(hook):
                    continue

                if not hook_is_async and asyncio.iscoroutinefunction(hook):
                    logger.error(
                        '%s.%s is a coroutine function',
                        middleware,
                        hook_name,
                    )

                self.hooks[hook_name].append(
                    (middleware, hook),
                )

                logger.debug('%s.%s discovered', middleware, hook_name)

    def _run_middlewares_sync(self, hook_name, data):
        if hook_name not in self.hooks:
            raise NotImplementedError(f"unknown hook '{hook_name}'")

        logger.debug('running %s with %r', hook_name, data)

        for middleware, hook in self.hooks[hook_name]:
            logger.debug('running %s.%s(%r)', middleware, hook_name, data)

            # run middleware hook
            return_value = hook(data)

            if return_value is None:
                # if the middleware does not return the data object it is
                # considered as handled

                logger.debug(
                    'data got handled by %s.%s', middleware, hook_name)

                return True, None, middleware

            elif return_value is not data:
                # if a middleware returns a custom value the data
                # is considered as handled

                logger.debug(
                    'data got handled by %s.%s by returning a custom value: %r',
                    middleware,
                    hook_name,
                    return_value,
                )

                return True, return_value, middleware

        logger.debug('data was not handled')

        return False, data, None

    async def _run_middlewares_async(self, hook_name, data):
        if hook_name not in self.hooks:
            raise NotImplementedError(f"unknown hook '{hook_name}'")

        logger.debug('running %s with %r', hook_name, data)

        for middleware, hook in self.hooks[hook_name]:
            logger.debug('running %s.%s(%r)', middleware, hook_name, data)

            try:
                await hook(data)

            except Exception:
                logger.exception('Exception raised while running %r', hook)

    async def run_on_startup(self, *args, **kwargs):
        data = MiddlewareData(
            server=self.server,
        )

        await self._run_middlewares_async(
            'on_startup',
            data,
        )

    async def run_on_shutdown(self, *args, **kwargs):
        data = MiddlewareData(
            server=self.server,
        )

        await self._run_middlewares_async(
            'on_shutdown',
            data,
        )

    async def handle_connection(self, connection):
        data = MiddlewareData(
            server=self.server,
            http_request=connection.http_request,
            connection=connection,
        )

        return await self.server.run_function_async(
            self._run_middlewares_sync,
            'handle_connection',
            data,
        )

    async def handle_websocket_message(self, connection, message):
        data = MiddlewareData(
            server=self.server,
            connection=connection,
            message=message,
        )

        return await self.server.run_function_async(
            self._run_middlewares_sync,
            'handle_websocket_message',
            data,
        )

    def handle_request(self, request, view):
        data = MiddlewareData(
            server=self.server,
            connection=request.connection,
            request=request,
            view=view,
        )

        return self._run_middlewares_sync(
            'handle_request',
            data,
        )



from threading import RLock


class Overlay:
    def __init__(self, server_state, data):
        self._server_state = server_state
        self._data = data

    @property
    def lock(self):
        return self._server_state.lock

    def __getattribute__(self, name):
        if name.startswith('_') or name == 'lock':
            return super().__getattribute__(name)

        attribute = self._data.__getattribute__(name)

        def shim(*args, **kwargs):
            with self.lock:
                return attribute(*args, **kwargs)

        if callable(attribute):
            return shim

        return attribute

    def __getitem__(self, key):
        with self.lock:
            item = self._data.__getitem__(key)

            if isinstance(item, (dict, list, Overlay)):
                return Overlay(self._server_state, item)

            return item

    def __setitem__(self, key, value):
        with self.lock:
            self._data.__setitem__(key, value)

    def __iter__(self):
        with self.lock:
            return self._data.__iter__()

    def __bool__(self):
        with self.lock:
            return self._data.__bool__()

    def __dir__(self):
        with self.lock:
            return self._data.__dir__()

    def __len__(self):
        with self.lock:
            return self._data.__len__()

    def __str__(self):
        with self.lock:
            return self._data.__str__()

    def __repr__(self):
        with self.lock:
            return self._data.__repr__()


class ServerState(Overlay):
    def __init__(self, initial_data=None):
        self._server_state = self
        self._data = initial_data or {}
        self._lock = RLock()

    @property
    def lock(self):
        return self._lock



import html

from lona.protocol import PATCH_TYPE, OPERATION

BOOLEAN_ATTRIBUTES = [
    'allowfullscreen',
    'async',
    'autofocus',
    'autoplay',
    'checked',
    'controls',
    'default',
    'disabled',
    'formnovalidate',
    'hidden',
    'ismap',
    'itemscope',
    'loop',
    'multiple',
    'muted',
    'novalidate',
    'open',
    'playsinline',
    'readonly',
    'required',
    'reversed',
    'selected',
]


class AttributeDict:
    PATCH_TYPE = PATCH_TYPE.ATTRIBUTES

    def __init__(self, node, *args, **kwargs):
        self._node = node
        self._attributes = dict(*args, **kwargs)

    # dict helper #############################################################
    def keys(self):
        with self._node.lock:
            return self._attributes.keys()

    def items(self):
        with self._node.lock:
            return self._attributes.items()

    def pop(self, name, *default):
        if len(default) > 1:
            raise TypeError(
                f'pop expected at most 2 arguments, got {1 + len(default)}',
            )

        with self._node.lock:
            if default and name not in self._attributes:
                return default[0]

            attribute = self._attributes.pop(name)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=self.PATCH_TYPE,
                operation=OPERATION.REMOVE,
                payload=[
                    name,
                ],
            )

            return attribute

    def clear(self):
        with self._node.lock:
            if not self._attributes:
                return

            self._attributes.clear()

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=self.PATCH_TYPE,
                operation=OPERATION.CLEAR,
                payload=[],
            )

    def get(self, *args, **kwargs):
        with self._node.lock:
            return self._attributes.get(*args, **kwargs)

    def update(self, value):
        if not isinstance(value, dict):
            raise ValueError('dict required')

        with self._node.lock:
            for key, value in value.items():
                self[key] = value

    def __getitem__(self, name):
        with self._node.lock:
            return self._attributes[name]

    def __setitem__(self, name, value, issuer=None):
        if not isinstance(value, (int, bool, float, str)):
            raise ValueError(f'unsupported type: {type(value)}')

        if name in ('id', 'class'):
            raise RuntimeError(
                f"Node.attributes['{name}'] is not supported. Use Node.{name}_list instead.",
            )

        if name == 'style':
            raise RuntimeError(
                f"Node.attributes['{name}'] is not supported. Use Node.{name} instead.",
            )

        with self._node.lock:
            if name in self._attributes and self._attributes[name] == value:
                return

            self._attributes[name] = value

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=self.PATCH_TYPE,
                operation=OPERATION.SET,
                payload=[
                    name,
                    value,
                ],
                issuer=issuer,
            )

    def __delitem__(self, name):
        self.pop(name, None)

    def __eq__(self, other):
        with self._node.lock:
            if isinstance(other, self.__class__):
                other = other._attributes

        return self._attributes == other

    def __iter__(self):
        with self._node.lock:
            return self._attributes.__iter__()

    def __contains__(self, name):
        with self._node.lock:
            return name in self._attributes

    def __bool__(self):
        with self._node.lock:
            return bool(self._attributes)

    # serialization ###########################################################
    def _reset(self, value):
        if not isinstance(value, dict):
            raise ValueError('unsupported type')

        for name, v in value.items():
            if not isinstance(v, (int, bool, float, str)):
                raise ValueError('unsupported type')

            if name in ('id', 'class'):
                raise RuntimeError(
                    f"Node.attributes['{name}'] is not supported. Use Node.{name}_list instead.",
                )

            if name == 'style':
                raise RuntimeError(
                    f"Node.attributes['{name}'] is not supported. Use Node.{name} instead.",
                )

        with self._node.lock:
            self._attributes = value

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=self.PATCH_TYPE,
                operation=OPERATION.RESET,
                payload=[
                    dict(value),
                ],
            )

    def _serialize(self):
        return dict(self._attributes)

    # string representation ###################################################
    def to_attribute_string(self, skip_value=False):
        with self._node.lock:
            string = []

            for key, value in self._attributes.items():
                if skip_value and key == 'value':
                    continue

                # boolean properties
                if key in BOOLEAN_ATTRIBUTES:
                    if value is not False:
                        string.append(key)

                else:
                    string.append(f'{key}="{value}"')

            return html.escape(' '.join(string))

    def to_sub_attribute_string(self):
        with self._node.lock:
            string = []

            for key, value in self._attributes.items():
                string.append(f'{key}: {value}')

            return html.escape('; '.join(string))

    def __repr__(self):
        return f'<AttributeDict({self._attributes!r})>'


class StyleDict(AttributeDict):
    PATCH_TYPE = PATCH_TYPE.STYLE

    def __repr__(self):
        return f'<StyleDict({self._attributes!r})>'



from tempfile import TemporaryDirectory
import logging
import os

from jinja2 import FileSystemLoader, Environment

from lona.protocol import ENUMS
from lona._json import dumps

SOURCE_ROOT = os.path.join(os.path.dirname(__file__), 'client')

logger = logging.getLogger('lona.client_pre_compiler')


class ClientPreCompiler:
    def __init__(self, server):
        self.server = server

        self.tmp_dir = TemporaryDirectory()

        self.jinja2_env = Environment(
            loader=FileSystemLoader(SOURCE_ROOT),
        )

        self.compile()

    def _get_path(self) -> str:
        return os.path.join(self.tmp_dir.name, 'lona.js')

    def get_settings(self):
        settings = self.server.settings

        return {
            'VIEW_START_TIMEOUT': settings.CLIENT_VIEW_START_TIMEOUT,
            'INPUT_EVENT_TIMEOUT': settings.CLIENT_INPUT_EVENT_TIMEOUT,
            'PING_INTERVAL': settings.CLIENT_PING_INTERVAL,
        }

    def get_enums(self):
        enums = {}

        for enum in ENUMS:
            enum_values = {}

            for enum_value in enum:
                enum_values[enum_value.name] = enum_value.value

            enums[enum.__name__] = enum_values

        return enums

    def compile(self):
        logger.debug('pre compiling client')

        try:
            path = self._get_path()
            template = self.jinja2_env.get_template('lona.js')

            template_context = {
                'protocol': dumps(self.get_enums()),
                'settings': dumps(self.get_settings()),
            }

            file_content = template.render(
                **template_context,
            )

            with open(path, 'w+') as f:
                f.write(file_content)
                f.close()

        except Exception:
            logger.exception('exception raised while pre compiling js client')

    def resolve(self) -> str:
        if self.server.settings.CLIENT_RECOMPILE:
            self.compile()

        return self._get_path()

    def __repr__(self):
        return f'<ClientPreCompiler({self._get_path()})>'
